from inspect import get_annotations
from typing import (
    List,
    Sequence,
    Dict,
    Generic,
    TypeVar,
    cast,
    overload,
    Any,
    TypeAlias,
    get_args,
    get_origin,
)
from collections import OrderedDict

from pyteal.types import TealType
from pyteal.errors import TealInputError, TealInternalError
from pyteal.ast.expr import Expr
from pyteal.ast.seq import Seq
from pyteal.ast.int import Int
from pyteal.ast.bytes import Bytes
from pyteal.ast.unaryexpr import Len
from pyteal.ast.binaryexpr import ExtractUint16
from pyteal.ast.naryexpr import Concat
from pyteal.ast.scratchvar import ScratchVar

from pyteal.ast.abi.type import TypeSpec, BaseType, ComputedValue
from pyteal.ast.abi.bool import (
    Bool,
    BoolTypeSpec,
    _consecutive_bool_instance_num,
    _consecutive_bool_type_spec_num,
    _bool_sequence_length,
    _encode_bool_sequence,
    _bool_aware_static_byte_length,
)
from pyteal.ast.abi.uint import NUM_BITS_IN_BYTE, Uint16
from pyteal.ast.abi.util import substring_for_decoding, type_spec_from_annotation


def _encode_tuple(values: Sequence[BaseType]) -> Expr:
    heads: List[Expr] = []
    head_length_static: int = 0

    dynamicValueIndexToHeadIndex: Dict[int, int] = dict()
    ignoreNext = 0
    for i, elem in enumerate(values):
        if ignoreNext > 0:
            ignoreNext -= 1
            continue

        elemType = elem.type_spec()

        if elemType == BoolTypeSpec():
            numBools = _consecutive_bool_instance_num(values, i)
            ignoreNext = numBools - 1
            head_length_static += _bool_sequence_length(numBools)
            heads.append(
                _encode_bool_sequence(cast(Sequence[Bool], values[i : i + numBools]))
            )
            continue

        if elemType.is_dynamic():
            head_length_static += 2
            dynamicValueIndexToHeadIndex[i] = len(heads)
            heads.append(Seq())  # a placeholder
            continue

        head_length_static += elemType.byte_length_static()
        heads.append(elem.encode())

    tail_offset = Uint16()
    tail_offset_accumulator = Uint16()
    tail_holder = ScratchVar(TealType.bytes)
    encoded_tail = ScratchVar(TealType.bytes)

    firstDynamicTail = True
    for i, elem in enumerate(values):
        if elem.type_spec().is_dynamic():
            if firstDynamicTail:
                firstDynamicTail = False
                updateVars = Seq(
                    tail_holder.store(encoded_tail.load()),
                    tail_offset.set(head_length_static),
                )
            else:
                updateVars = Seq(
                    tail_holder.store(Concat(tail_holder.load(), encoded_tail.load())),
                    tail_offset.set(tail_offset_accumulator),
                )

            notLastDynamicValue = any(
                [nextValue.type_spec().is_dynamic() for nextValue in values[i + 1 :]]
            )
            if notLastDynamicValue:
                updateAccumulator = tail_offset_accumulator.set(
                    tail_offset.get() + Len(encoded_tail.load())
                )
            else:
                updateAccumulator = Seq()

            heads[dynamicValueIndexToHeadIndex[i]] = Seq(
                encoded_tail.store(elem.encode()),
                updateVars,
                updateAccumulator,
                tail_offset.encode(),
            )

    toConcat = heads
    if not firstDynamicTail:
        toConcat.append(tail_holder.load())

    if len(toConcat) == 0:
        return Bytes("")

    return Concat(*toConcat)


def _index_tuple(
    value_types: Sequence[TypeSpec], encoded: Expr, index: int, output: BaseType
) -> Expr:
    if not (0 <= index < len(value_types)):
        raise ValueError("Index outside of range")

    offset = 0
    ignoreNext = 0
    lastBoolStart = 0
    lastBoolLength = 0
    for i, typeBefore in enumerate(value_types[:index]):
        if ignoreNext > 0:
            ignoreNext -= 1
            continue

        if typeBefore == BoolTypeSpec():
            lastBoolStart = offset
            lastBoolLength = _consecutive_bool_type_spec_num(value_types, i)
            offset += _bool_sequence_length(lastBoolLength)
            ignoreNext = lastBoolLength - 1
            continue

        if typeBefore.is_dynamic():
            offset += 2
            continue

        offset += typeBefore.byte_length_static()

    valueType = value_types[index]
    if output.type_spec() != valueType:
        raise TypeError("Output type does not match value type")

    if type(output) is Bool:
        if ignoreNext > 0:
            # value is in the middle of a bool sequence
            bitOffsetInBoolSeq = lastBoolLength - ignoreNext
            bitOffsetInEncoded = lastBoolStart * NUM_BITS_IN_BYTE + bitOffsetInBoolSeq
        else:
            # value is the beginning of a bool sequence (or a single bool)
            bitOffsetInEncoded = offset * NUM_BITS_IN_BYTE
        return output.decode_bit(encoded, Int(bitOffsetInEncoded))

    if valueType.is_dynamic():
        hasNextDynamicValue = False
        nextDynamicValueOffset = offset + 2
        ignoreNext = 0
        for i, typeAfter in enumerate(value_types[index + 1 :], start=index + 1):
            if ignoreNext > 0:
                ignoreNext -= 1
                continue

            if type(typeAfter) is BoolTypeSpec:
                boolLength = _consecutive_bool_type_spec_num(value_types, i)
                nextDynamicValueOffset += _bool_sequence_length(boolLength)
                ignoreNext = boolLength - 1
                continue

            if typeAfter.is_dynamic():
                hasNextDynamicValue = True
                break

            nextDynamicValueOffset += typeAfter.byte_length_static()

        start_index = ExtractUint16(encoded, Int(offset))
        if not hasNextDynamicValue:
            # This is the final dynamic value, so decode the substring from start_index to the end of
            # encoded
            return output.decode(encoded, start_index=start_index)

        # There is a dynamic value after this one, and end_index is where its tail starts, so decode
        # the substring from start_index to end_index
        end_index = ExtractUint16(encoded, Int(nextDynamicValueOffset))
        return output.decode(encoded, start_index=start_index, end_index=end_index)

    start_index = Int(offset)
    length = Int(valueType.byte_length_static())

    if index + 1 == len(value_types):
        if offset == 0:
            # This is the first and only value in the tuple, so decode all of encoded
            return output.decode(encoded)
        # This is the last value in the tuple, so decode the substring from start_index to the end of
        # encoded
        return output.decode(encoded, start_index=start_index)

    if offset == 0:
        # This is the first value in the tuple, so decode the substring from 0 with length length
        return output.decode(encoded, length=length)

    # This is not the first or last value, so decode the substring from start_index with length length
    return output.decode(encoded, start_index=start_index, length=length)


class TupleTypeSpec(TypeSpec):
    def __init__(self, *value_type_specs: TypeSpec) -> None:
        super().__init__()
        self.value_specs = list(value_type_specs)

    def value_type_specs(self) -> List[TypeSpec]:
        """Get the TypeSpecs for the values of this tuple."""
        return self.value_specs

    def length_static(self) -> int:
        """Get the number of values this tuple holds."""
        return len(self.value_specs)

    def new_instance(self) -> "Tuple":
        match self.length_static():
            case 0:
                return Tuple0()
            case 1:
                return Tuple1(self)
            case 2:
                return Tuple2(self)
            case 3:
                return Tuple3(self)
            case 4:
                return Tuple4(self)
            case 5:
                return Tuple5(self)
            case _:
                return Tuple(self)

    def annotation_type(self) -> "type[Tuple]":
        vtses = self.value_type_specs()

        def annotater():
            return [x.annotation_type() for x in vtses]

        match len(vtses):
            case 0:
                return Tuple0
            case 1:
                v0 = annotater()[0]
                return Tuple1[v0]  # type: ignore[valid-type]
            case 2:
                v0, v1 = annotater()
                return Tuple2[v0, v1]  # type: ignore[valid-type]
            case 3:
                v0, v1, v2 = annotater()
                return Tuple3[v0, v1, v2]  # type: ignore[valid-type]
            case 4:
                v0, v1, v2, v3 = annotater()
                return Tuple4[v0, v1, v2, v3]  # type: ignore[valid-type]
            case 5:
                v0, v1, v2, v3, v4 = annotater()
                return Tuple5[v0, v1, v2, v3, v4]  # type: ignore[valid-type]

        raise TypeError(f"Cannot annotate tuple of length {len(vtses)}")

    def is_dynamic(self) -> bool:
        return any(type_spec.is_dynamic() for type_spec in self.value_type_specs())

    def byte_length_static(self) -> int:
        if self.is_dynamic():
            raise ValueError("Type is dynamic")
        return _bool_aware_static_byte_length(self.value_type_specs())

    def storage_type(self) -> TealType:
        return TealType.bytes

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, TupleTypeSpec)
            and self.value_type_specs() == other.value_type_specs()
        )

    def __str__(self) -> str:
        return "({})".format(",".join(map(str, self.value_type_specs())))


TupleTypeSpec.__module__ = "pyteal.abi"


class Tuple(BaseType):
    def __init__(self, tuple_type_spec: TupleTypeSpec) -> None:
        super().__init__(tuple_type_spec)

    def type_spec(self) -> TupleTypeSpec:
        return cast(TupleTypeSpec, super().type_spec())

    def decode(
        self,
        encoded: Expr,
        *,
        start_index: Expr = None,
        end_index: Expr = None,
        length: Expr = None,
    ) -> Expr:
        extracted = substring_for_decoding(
            encoded, start_index=start_index, end_index=end_index, length=length
        )
        return self.stored_value.store(extracted)

    @overload
    def set(self, *values: BaseType) -> Expr:
        pass

    @overload
    def set(self, values: ComputedValue["Tuple"]) -> Expr:
        # TODO: should support values as a Tuple as well
        pass

    def set(self, *values):
        """
        set(*values: BaseType) -> pyteal.Expr
        set(values: ComputedValue[Tuple]) -> pyteal.Expr

        Set the elements of this Tuple to the input values.

        The behavior of this method depends on the input argument type:

            * Variable number of :code:`BaseType` arguments: set the elements of this Tuple to the arguments to this method. A compiler error will occur if any argument does not match this Tuple's element type at the same index, or if the total argument count does not equal this Tuple's length.
            * :code:`ComputedValue[Tuple]`: copy the elements from a Tuple produced by a ComputedValue. The element types and length produced by the ComputedValue must exactly match this Tuple's element types and length, otherwise an error will occur.

        Args:
            values: The new elements this Tuple should have. This must follow the above constraints.

        Returns:
            An expression which stores the given value into this Tuple.
        """
        if len(values) == 1 and isinstance(values[0], ComputedValue):
            return self._set_with_computed_type(values[0])

        for value in values:
            if not isinstance(value, BaseType):
                raise TealInputError(f"Expected BaseType, got {value}")

        myTypes = self.type_spec().value_type_specs()
        if len(myTypes) != len(values):
            raise TealInputError(
                f"Incorrect length for values. Expected {len(myTypes)}, got {len(values)}"
            )
        if not all(myTypes[i] == values[i].type_spec() for i in range(len(myTypes))):
            raise TealInputError("Input values do not match type")
        return self.stored_value.store(_encode_tuple(values))

    def encode(self) -> Expr:
        return self.stored_value.load()

    def length(self) -> Expr:
        """Get the number of values this tuple holds as an Expr."""
        return Int(self.type_spec().length_static())

    def __getitem__(self, index: int) -> "TupleElement[Any]":
        """Retrieve an element by its index in this Tuple.

        Indexes start at 0.

        Args:
            index: a Python integer containing the index to access. This function will raise an error
                if its value is negative or if the index is equal to or greater than the length of
                this Tuple.

        Returns:
            A TupleElement that corresponds to the element at the given index. This type is a
            ComputedValue. Due to Python type limitations, the parameterized type of the
            TupleElement is Any.
        """
        if not (0 <= index < self.type_spec().length_static()):
            raise TealInputError(f"Index out of bounds: {index}")
        return TupleElement(self, index)


Tuple.__module__ = "pyteal.abi"

T = TypeVar("T", bound=BaseType)


class TupleElement(ComputedValue[T]):
    """Represents the extraction of a specific element from a Tuple."""

    def __init__(self, tuple: Tuple, index: int) -> None:
        super().__init__()
        self.tuple = tuple
        self.index = index

    def produced_type_spec(self) -> TypeSpec:
        return self.tuple.type_spec().value_type_specs()[self.index]

    def store_into(self, output: T) -> Expr:
        return _index_tuple(
            self.tuple.type_spec().value_type_specs(),
            self.tuple.encode(),
            self.index,
            output,
        )


TupleElement.__module__ = "pyteal.abi"

# Until Python 3.11 is released with support for PEP 646 -- Variadic Generics, it's not possible for
# the Tuple class to take an arbitrary number of template parameters. As a workaround, we define the
# following classes for specifically sized Tuples. If needed, more classes can be added for larger
# sizes.


def _tuple_raise_arg_mismatch(expected: int, typespec: TupleTypeSpec):
    if len(typespec.value_specs) != expected:
        raise TealInputError(
            f"Expected TupleTypeSpec with {expected} elements, Got {len(typespec.value_specs)}"
        )


class Tuple0(Tuple):
    """A Tuple with 0 values."""

    def __init__(self) -> None:
        super().__init__(TupleTypeSpec())


Tuple0.__module__ = "pyteal.abi"

T1 = TypeVar("T1", bound=BaseType)


class Tuple1(Tuple, Generic[T1]):
    """A Tuple with 1 value."""

    def __init__(self, value_type_spec: TupleTypeSpec) -> None:
        _tuple_raise_arg_mismatch(1, value_type_spec)
        super().__init__(value_type_spec)


Tuple1.__module__ = "pyteal.abi"

T2 = TypeVar("T2", bound=BaseType)


class Tuple2(Tuple, Generic[T1, T2]):
    """A Tuple with 2 values."""

    def __init__(self, value_type_spec: TupleTypeSpec) -> None:
        _tuple_raise_arg_mismatch(2, value_type_spec)
        super().__init__(value_type_spec)


Tuple2.__module__ = "pyteal.abi"

T3 = TypeVar("T3", bound=BaseType)


class Tuple3(Tuple, Generic[T1, T2, T3]):
    """A Tuple with 3 values."""

    def __init__(
        self,
        value_type_spec: TupleTypeSpec,
    ) -> None:
        _tuple_raise_arg_mismatch(3, value_type_spec)
        super().__init__(value_type_spec)


Tuple3.__module__ = "pyteal.abi"

T4 = TypeVar("T4", bound=BaseType)


class Tuple4(Tuple, Generic[T1, T2, T3, T4]):
    """A Tuple with 4 values."""

    def __init__(
        self,
        value_type_spec: TupleTypeSpec,
    ) -> None:
        _tuple_raise_arg_mismatch(4, value_type_spec)
        super().__init__(value_type_spec)


Tuple4.__module__ = "pyteal.abi"

T5 = TypeVar("T5", bound=BaseType)


class Tuple5(Tuple, Generic[T1, T2, T3, T4, T5]):
    """A Tuple with 5 values."""

    def __init__(
        self,
        value_type_spec: TupleTypeSpec,
    ) -> None:
        _tuple_raise_arg_mismatch(5, value_type_spec)
        super().__init__(value_type_spec)


Tuple5.__module__ = "pyteal.abi"


Field: TypeAlias = TupleElement[T]


class NamedTupleTypeSpec(TupleTypeSpec):
    """A NamedTupleType inherits from TupleTypeSpec, allowing for more than 5 elements."""

    def __init__(
        self, instance_class: type["NamedTuple"], *value_type_specs: TypeSpec
    ) -> None:
        if instance_class == NamedTuple:
            raise TealInputError(
                "NamedTupleTypeSpec must be instanced with subclassed NamedTuple class."
            )

        self.instance_class: type["NamedTuple"] = instance_class
        super().__init__(*value_type_specs)

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, NamedTupleTypeSpec)
            and self.instance_class == other.instance_class
            and self.value_type_specs() == other.value_type_specs()
        )

    def annotation_type(self) -> "type[NamedTuple]":
        return self.instance_class

    def new_instance(self) -> "NamedTuple":
        return self.instance_class()


NamedTupleTypeSpec.__module__ = "pyteal.abi"


class NamedTuple(Tuple):
    """A NamedTuple is a :any:`Tuple` that has named elements, inspired by Python's `typing.NamedTuple <https://docs.python.org/3/library/typing.html#typing.NamedTuple>`_.

    A new NamedTuple type can be created by subclassing this class and adding field annotations.
    Every field annotation must be an instantiable ABI type wrapped in the :code:`abi.Field` annotation.

    For example:

        .. code-block:: python

            from pyteal import *

            class User(abi.NamedTuple):
                address: abi.Field[abi.Address]
                balance: abi.Field[abi.Uint64]

            # User is equivalent to abi.Tuple2[abi.Address, abi.Uint64]

            my_user = User()

    .. automethod:: __init_subclass__
    .. automethod:: __getattr__
    """

    def __init_subclass__(cls) -> None:
        """This method ensures one only constructs directly from `NamedTuple`,
        rather than inheriting from `NamedTuple`'s inheritance.

        We want to prohibit the following examples:

            .. code-block:: python

                from pyteal import *

                class LegalInheritance(abi.NamedTuple):
                    a: abi.Field[abi.Uint64]

                # following are bad cases we guard against

                class IllegalInheritance0(LegalInheritance):
                    a: abi.Field[abi.Uint64]

                class IllegalInheritance1(LegalInheritance, abi.NamedTuple):
                    a: abi.Field[abi.Uint64]
        """
        is_named_tuple_in_bases = False

        for base_t in cls.__bases__:
            if base_t == NamedTuple:
                is_named_tuple_in_bases = True
            elif issubclass(base_t, NamedTuple):
                raise TealInternalError(
                    f"Cannot construct {cls} by inheriting {cls.__bases__}. "
                    f"Must be constructed by direct inheritance from NamedTuple"
                )

        if not is_named_tuple_in_bases:
            raise TealInternalError(
                "Unexpected: did not find NamedTuple in __bases__,"
                "did not find NamedTuple in __bases__ member's __bases__"
            )

        super().__init_subclass__()

    def __init__(self):
        if type(self) is NamedTuple:
            raise TealInputError("NamedTuple must be subclassed")

        anns = get_annotations(type(self))
        if not anns:
            raise TealInputError("Expected fields to be declared but found none")

        # NOTE: this `_ready` variable enables `__setattr__` during `__init__` execution,
        # while after `__init__`, we cannot use `__setattr__` to set fields in `NamedTuple`.
        # NOTE: If we declare variable `__ready`, then internally,
        # the variable name would be changed to `_NamedTuple__ready`, which is implicit.
        self.__ready = False
        self.__type_specs: OrderedDict[str, TypeSpec] = OrderedDict()
        self.__field_index: dict[str, int] = {}

        for index, (name, annotation) in enumerate(anns.items()):
            origin = get_origin(annotation)
            if origin is None:
                origin = annotation
            if origin is not get_origin(Field):
                raise TealInputError(
                    f'Type annotation for attribute "{name}" must be a Field. Got {origin}'
                )

            args = get_args(annotation)
            if len(args) != 1:
                raise TealInputError(
                    f'Type annotation for attribute "{name}" must have a single argument. Got {args}'
                )

            self.__type_specs[name] = type_spec_from_annotation(args[0])
            self.__field_index[name] = index

        super().__init__(
            NamedTupleTypeSpec(type(self), *list(self.__type_specs.values()))
        )

        self.__ready = True

    def __getattr__(self, field: str) -> TupleElement[Any]:
        """Retrieve an element by its field in this NamedTuple.

        For example:

        .. code-block:: python

            from pyteal import *

            class User(abi.NamedTuple):
                address: abi.Field[abi.Address]
                balance: abi.Field[abi.Uint64]

            @ABIReturnSubroutine
            def get_user_balance(user: User, *, output: abi.Uint64) -> Expr:
                return output.set(user.balance)

        Args:
            field: a Python string containing the field to access.
                This function will raise an KeyError if the field is not available in the defined NamedTuple.

        Returns:
            A TupleElement that corresponds to the element at the given field name, returning a ComputedType.
            Due to Python type limitations, the parameterized type of the TupleElement is Any.
        """
        return self.__getitem__(self.__field_index[field])

    def __setattr__(self, name: str, field: Any) -> None:
        # we allow `__setattr__` only when:
        # - we are in `__init__`: `not self.__ready`
        # - we are setting `_ready`: `name == "_NamedTuple__ready"`,
        #   it is internally changed from `__ready` to `_NamedTuple__ready`,
        #   see notes in `__init__`
        if name == "_NamedTuple__ready" or not self.__ready:
            super().__setattr__(name, field)
            return
        raise TealInputError("cannot assign to NamedTuple attributes.")


NamedTuple.__module__ = "pyteal.abi"



from typing import Optional, List

from pyteal.ir.tealop import TealOp
from pyteal.ir.tealblock import TealBlock


class TealConditionalBlock(TealBlock):
    """Represents a basic block of TealComponents in a graph ending with a branch condition."""

    def __init__(self, ops: List[TealOp]) -> None:
        super().__init__(ops)
        self.trueBlock: Optional[TealBlock] = None
        self.falseBlock: Optional[TealBlock] = None

    def setTrueBlock(self, block: TealBlock) -> None:
        """Set the block that this one should branch to if its condition is true."""
        self.trueBlock = block

    def setFalseBlock(self, block: TealBlock) -> None:
        """Set the block that this one should branch to if its condition is false."""
        self.falseBlock = block

    def getOutgoing(self) -> List[TealBlock]:
        outgoing = []
        if self.trueBlock is not None:
            outgoing.append(self.trueBlock)
        if self.falseBlock is not None:
            outgoing.append(self.falseBlock)
        return outgoing

    def replaceOutgoing(self, oldBlock: TealBlock, newBlock: TealBlock) -> None:
        if self.trueBlock is oldBlock:
            self.trueBlock = newBlock
        elif self.falseBlock is oldBlock:
            self.falseBlock = newBlock

    def __repr__(self) -> str:
        return "TealConditionalBlock({}, true={}, false={})".format(
            repr(self.ops),
            repr(self.trueBlock),
            repr(self.falseBlock),
        )

    def __eq__(self, other: object) -> bool:
        if type(other) is not TealConditionalBlock:
            return False
        return (
            self.ops == other.ops
            and self.trueBlock == other.trueBlock
            and self.falseBlock == other.falseBlock
        )


TealConditionalBlock.__module__ = "pyteal"



import pytest

import pyteal as pt
from pyteal import abi


POSITIVE_CASES = [
    abi.Uint16(),
    abi.Uint32(),
    abi.StaticArray(abi.StaticArrayTypeSpec(abi.BoolTypeSpec(), 12)),
]


@pytest.mark.parametrize("case", POSITIVE_CASES)
def test_method_return(case):
    m_ret = abi.MethodReturn(case)
    assert m_ret.type_of() == pt.TealType.none
    assert not m_ret.has_return()
    assert str(m_ret) == f"(MethodReturn {case.type_spec()})"


NEGATIVE_CASES = [
    pt.Int(0),
    pt.Bytes("aaaaaaa"),
    abi.Uint16,
    abi.Uint32,
]


@pytest.mark.parametrize("case", NEGATIVE_CASES)
def test_method_return_error(case):
    with pytest.raises(pt.TealInputError):
        abi.MethodReturn(case)



import pytest

import pyteal as pt

options = pt.CompileOptions()


def test_and_one():
    arg = pt.Int(1)
    expr = pt.And(arg)
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock([pt.TealOp(arg, pt.Op.int, 1)])

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_and_two():
    args = [pt.Int(1), pt.Int(2)]
    expr = pt.And(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 2),
            pt.TealOp(expr, pt.Op.logic_and),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_and_three():
    args = [pt.Int(1), pt.Int(2), pt.Int(3)]
    expr = pt.And(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 2),
            pt.TealOp(expr, pt.Op.logic_and),
            pt.TealOp(args[2], pt.Op.int, 3),
            pt.TealOp(expr, pt.Op.logic_and),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_and_overload():
    args = [pt.Int(1), pt.Int(2)]
    expr = args[0].And(args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 2),
            pt.TealOp(expr, pt.Op.logic_and),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_and_invalid():
    with pytest.raises(pt.TealInputError):
        pt.And()

    with pytest.raises(pt.TealTypeError):
        pt.And(pt.Int(1), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.And(pt.Txn.receiver(), pt.Int(1))

    with pytest.raises(pt.TealTypeError):
        pt.And(pt.Txn.receiver(), pt.Txn.receiver())


def test_or_one():
    arg = pt.Int(1)
    expr = pt.Or(arg)
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock([pt.TealOp(arg, pt.Op.int, 1)])

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_or_two():
    args = [pt.Int(1), pt.Int(0)]
    expr = pt.Or(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.logic_or),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_or_three():
    args = [pt.Int(0), pt.Int(1), pt.Int(2)]
    expr = pt.Or(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 0),
            pt.TealOp(args[1], pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.logic_or),
            pt.TealOp(args[2], pt.Op.int, 2),
            pt.TealOp(expr, pt.Op.logic_or),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_or_overload():
    args = [pt.Int(1), pt.Int(0)]
    expr = args[0].Or(args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.logic_or),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_or_invalid():
    with pytest.raises(pt.TealInputError):
        pt.Or()

    with pytest.raises(pt.TealTypeError):
        pt.Or(pt.Int(1), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.Or(pt.Txn.receiver(), pt.Int(1))

    with pytest.raises(pt.TealTypeError):
        pt.Or(pt.Txn.receiver(), pt.Txn.receiver())


def test_concat_one():
    arg = pt.Bytes("a")
    expr = pt.Concat(arg)
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock([pt.TealOp(arg, pt.Op.byte, '"a"')])

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_concat_two():
    args = [pt.Bytes("a"), pt.Bytes("b")]
    expr = pt.Concat(args[0], args[1])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"a"'),
            pt.TealOp(args[1], pt.Op.byte, '"b"'),
            pt.TealOp(expr, pt.Op.concat),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_concat_three():
    args = [pt.Bytes("a"), pt.Bytes("b"), pt.Bytes("c")]
    expr = pt.Concat(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"a"'),
            pt.TealOp(args[1], pt.Op.byte, '"b"'),
            pt.TealOp(expr, pt.Op.concat),
            pt.TealOp(args[2], pt.Op.byte, '"c"'),
            pt.TealOp(expr, pt.Op.concat),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_concat_invalid():
    with pytest.raises(pt.TealInputError):
        pt.Concat()

    with pytest.raises(pt.TealTypeError):
        pt.Concat(pt.Int(1), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.Concat(pt.Txn.receiver(), pt.Int(1))

    with pytest.raises(pt.TealTypeError):
        pt.Concat(pt.Int(1), pt.Int(2))



from typing import Callable, NamedTuple, Literal, Optional, Any, get_origin
from inspect import isabstract
import pytest

import algosdk.abi

import pyteal as pt
from pyteal import abi
from pyteal.ast.abi.util import (
    substring_for_decoding,
    int_literal_from_annotation,
    type_spec_from_algosdk,
    type_spec_is_assignable_to,
)

options = pt.CompileOptions(version=5)


def bfs_on_inheritance(tt: type):
    type_queue = [tt]
    visited_types = set()
    while type_queue:
        current_type = type_queue.pop()
        yield current_type
        visited_types.add(current_type)
        for child_type in current_type.__subclasses__():
            if child_type in visited_types:
                continue
            type_queue.append(child_type)


class SubstringTest(NamedTuple):
    start_index: Optional[pt.Expr]
    end_index: Optional[pt.Expr]
    length: Optional[pt.Expr]
    expected: Callable[[pt.Expr], pt.Expr | type[Exception]]


SUBSTRING_TEST_CASES: list[SubstringTest] = [
    SubstringTest(
        start_index=None, end_index=None, length=None, expected=lambda encoded: encoded
    ),
    SubstringTest(
        start_index=None,
        end_index=None,
        length=pt.Int(4),
        expected=lambda encoded: pt.Extract(encoded, pt.Int(0), pt.Int(4)),
    ),
    SubstringTest(
        start_index=None,
        end_index=pt.Int(4),
        length=None,
        expected=lambda encoded: pt.Substring(encoded, pt.Int(0), pt.Int(4)),
    ),
    SubstringTest(
        start_index=None,
        end_index=pt.Int(4),
        length=pt.Int(5),
        expected=lambda _: pt.TealInputError,
    ),
    SubstringTest(
        start_index=pt.Int(4),
        end_index=None,
        length=None,
        expected=lambda encoded: pt.Suffix(encoded, pt.Int(4)),
    ),
    SubstringTest(
        start_index=pt.Int(4),
        end_index=None,
        length=pt.Int(5),
        expected=lambda encoded: pt.Extract(encoded, pt.Int(4), pt.Int(5)),
    ),
    SubstringTest(
        start_index=pt.Int(4),
        end_index=pt.Int(5),
        length=None,
        expected=lambda encoded: pt.Substring(encoded, pt.Int(4), pt.Int(5)),
    ),
    SubstringTest(
        start_index=pt.Int(4),
        end_index=pt.Int(5),
        length=pt.Int(6),
        expected=lambda _: pt.TealInputError,
    ),
]


@pytest.mark.parametrize(
    "start_index, end_index, length, expected", SUBSTRING_TEST_CASES
)
def test_substringForDecoding(
    start_index: Optional[pt.Expr],
    end_index: Optional[pt.Expr],
    length: Optional[pt.Expr],
    expected: Callable[[pt.Expr], pt.Expr | type[Exception]],
):
    encoded = pt.Bytes("encoded")

    expected_expr = expected(encoded)

    if not isinstance(expected_expr, pt.Expr):
        with pytest.raises(expected_expr):
            substring_for_decoding(
                encoded,
                start_index=start_index,
                end_index=end_index,
                length=length,
            )
        return

    expr = substring_for_decoding(
        encoded,
        start_index=start_index,
        end_index=end_index,
        length=length,
    )
    assert expr.type_of() == pt.TealType.bytes
    assert not expr.has_return()

    expected_blocks, _ = expected_expr.__teal__(options)
    expected_blocks.addIncoming()
    expected_blocks = pt.TealBlock.NormalizeBlocks(expected_blocks)

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected_blocks


class IntAnnotationTest(NamedTuple):
    annotation: Any
    expected: int | type[Exception]


IN_ANNOTATION_TEST_CASES: list[IntAnnotationTest] = [
    IntAnnotationTest(annotation=Literal[0], expected=0),
    IntAnnotationTest(annotation=Literal[1], expected=1),
    IntAnnotationTest(annotation=Literal[10], expected=10),
    IntAnnotationTest(annotation=Literal[True], expected=TypeError),
    IntAnnotationTest(annotation=Literal["test"], expected=TypeError),
    IntAnnotationTest(annotation=Literal[b"test"], expected=TypeError),
    IntAnnotationTest(annotation=Literal[None], expected=TypeError),
    IntAnnotationTest(annotation=Literal[0, 1], expected=TypeError),
    IntAnnotationTest(annotation=Literal, expected=TypeError),
]


@pytest.mark.parametrize("annotation, expected", IN_ANNOTATION_TEST_CASES)
def test_int_literal_from_annotation(annotation: Any, expected: int | type[Exception]):
    if not isinstance(expected, int):
        with pytest.raises(expected):
            int_literal_from_annotation(annotation)
        return

    actual = int_literal_from_annotation(annotation)
    assert actual == expected


class TypeAnnotationTest(NamedTuple):
    annotation: Any
    expected: abi.TypeSpec | type[Exception]


class ExampleNamedTuple(abi.NamedTuple):
    a: abi.Field[abi.Uint16]
    b: abi.Field[abi.DynamicArray[abi.Byte]]
    c: abi.Field[abi.Address]


TYPE_ANNOTATION_TEST_CASES: list[TypeAnnotationTest] = [
    TypeAnnotationTest(annotation=abi.Bool, expected=abi.BoolTypeSpec()),
    TypeAnnotationTest(annotation=abi.Byte, expected=abi.ByteTypeSpec()),
    TypeAnnotationTest(annotation=abi.Uint8, expected=abi.Uint8TypeSpec()),
    TypeAnnotationTest(annotation=abi.Uint16, expected=abi.Uint16TypeSpec()),
    TypeAnnotationTest(annotation=abi.Uint32, expected=abi.Uint32TypeSpec()),
    TypeAnnotationTest(annotation=abi.Uint64, expected=abi.Uint64TypeSpec()),
    TypeAnnotationTest(
        annotation=abi.DynamicArray[abi.Uint32],
        expected=abi.DynamicArrayTypeSpec(abi.Uint32TypeSpec()),
    ),
    TypeAnnotationTest(
        annotation=abi.DynamicArray[abi.Uint64],
        expected=abi.DynamicArrayTypeSpec(abi.Uint64TypeSpec()),
    ),
    TypeAnnotationTest(
        annotation=abi.DynamicArray[abi.DynamicArray[abi.Uint32]],
        expected=abi.DynamicArrayTypeSpec(
            abi.DynamicArrayTypeSpec(abi.Uint32TypeSpec())
        ),
    ),
    TypeAnnotationTest(
        annotation=abi.DynamicArray,
        expected=TypeError,
    ),
    TypeAnnotationTest(
        annotation=abi.StaticArray[abi.Uint32, Literal[0]],
        expected=abi.StaticArrayTypeSpec(abi.Uint32TypeSpec(), 0),
    ),
    TypeAnnotationTest(
        annotation=abi.StaticArray[abi.Uint32, Literal[10]],
        expected=abi.StaticArrayTypeSpec(abi.Uint32TypeSpec(), 10),
    ),
    TypeAnnotationTest(
        annotation=abi.StaticArray[abi.Bool, Literal[500]],
        expected=abi.StaticArrayTypeSpec(abi.BoolTypeSpec(), 500),
    ),
    TypeAnnotationTest(
        annotation=abi.StaticArray[abi.Bool, Literal[-1]],
        expected=TypeError,
    ),
    TypeAnnotationTest(
        annotation=abi.StaticArray[abi.Bool, int],
        expected=TypeError,
    ),
    TypeAnnotationTest(
        annotation=abi.StaticArray,
        expected=TypeError,
    ),
    TypeAnnotationTest(
        annotation=abi.StaticArray[abi.StaticArray[abi.Bool, Literal[500]], Literal[5]],
        expected=abi.StaticArrayTypeSpec(
            abi.StaticArrayTypeSpec(abi.BoolTypeSpec(), 500), 5
        ),
    ),
    TypeAnnotationTest(annotation=abi.Address, expected=abi.AddressTypeSpec()),
    TypeAnnotationTest(
        annotation=abi.StaticBytes[Literal[10]],
        expected=abi.StaticBytesTypeSpec(10),
    ),
    TypeAnnotationTest(
        annotation=abi.DynamicArray[abi.StaticArray[abi.Bool, Literal[500]]],
        expected=abi.DynamicArrayTypeSpec(
            abi.StaticArrayTypeSpec(abi.BoolTypeSpec(), 500)
        ),
    ),
    TypeAnnotationTest(annotation=abi.String, expected=abi.StringTypeSpec()),
    TypeAnnotationTest(
        annotation=abi.DynamicBytes,
        expected=abi.DynamicBytesTypeSpec(),
    ),
    TypeAnnotationTest(annotation=abi.Tuple, expected=abi.TupleTypeSpec()),
    TypeAnnotationTest(annotation=abi.Tuple0, expected=abi.TupleTypeSpec()),
    TypeAnnotationTest(
        annotation=abi.Tuple1[abi.Uint32],
        expected=abi.TupleTypeSpec(abi.Uint32TypeSpec()),
    ),
    TypeAnnotationTest(
        annotation=abi.Tuple1,
        expected=TypeError,
    ),
    TypeAnnotationTest(
        annotation=abi.Tuple2[abi.Uint32, abi.Uint16],
        expected=abi.TupleTypeSpec(abi.Uint32TypeSpec(), abi.Uint16TypeSpec()),
    ),
    TypeAnnotationTest(
        annotation=abi.Tuple2,
        expected=TypeError,
    ),
    TypeAnnotationTest(
        annotation=abi.Tuple3[abi.Uint32, abi.Uint16, abi.Byte],
        expected=abi.TupleTypeSpec(
            abi.Uint32TypeSpec(), abi.Uint16TypeSpec(), abi.ByteTypeSpec()
        ),
    ),
    TypeAnnotationTest(
        annotation=abi.Tuple3,
        expected=TypeError,
    ),
    TypeAnnotationTest(
        annotation=abi.Tuple3[
            abi.Tuple1[abi.Uint32],
            abi.StaticArray[abi.Bool, Literal[55]],
            abi.Tuple2[abi.Uint32, abi.Uint16],
        ],
        expected=abi.TupleTypeSpec(
            abi.TupleTypeSpec(abi.Uint32TypeSpec()),
            abi.StaticArrayTypeSpec(abi.BoolTypeSpec(), 55),
            abi.TupleTypeSpec(abi.Uint32TypeSpec(), abi.Uint16TypeSpec()),
        ),
    ),
    TypeAnnotationTest(
        annotation=abi.Tuple4[abi.Uint32, abi.Uint16, abi.Byte, abi.Bool],
        expected=abi.TupleTypeSpec(
            abi.Uint32TypeSpec(),
            abi.Uint16TypeSpec(),
            abi.ByteTypeSpec(),
            abi.BoolTypeSpec(),
        ),
    ),
    TypeAnnotationTest(
        annotation=abi.Tuple4,
        expected=TypeError,
    ),
    TypeAnnotationTest(
        annotation=abi.Tuple5[abi.Uint32, abi.Uint16, abi.Byte, abi.Bool, abi.Tuple0],
        expected=abi.TupleTypeSpec(
            abi.Uint32TypeSpec(),
            abi.Uint16TypeSpec(),
            abi.ByteTypeSpec(),
            abi.BoolTypeSpec(),
            abi.TupleTypeSpec(),
        ),
    ),
    TypeAnnotationTest(
        annotation=abi.Tuple5,
        expected=TypeError,
    ),
    TypeAnnotationTest(
        annotation=ExampleNamedTuple, expected=ExampleNamedTuple().type_spec()
    ),
    TypeAnnotationTest(
        annotation=list[abi.Uint16],
        expected=TypeError,
    ),
    TypeAnnotationTest(annotation=abi.Transaction, expected=abi.TransactionTypeSpec()),
    TypeAnnotationTest(
        annotation=abi.PaymentTransaction, expected=abi.PaymentTransactionTypeSpec()
    ),
    TypeAnnotationTest(
        annotation=abi.KeyRegisterTransaction,
        expected=abi.KeyRegisterTransactionTypeSpec(),
    ),
    TypeAnnotationTest(
        annotation=abi.AssetConfigTransaction,
        expected=abi.AssetConfigTransactionTypeSpec(),
    ),
    TypeAnnotationTest(
        annotation=abi.AssetFreezeTransaction,
        expected=abi.AssetFreezeTransactionTypeSpec(),
    ),
    TypeAnnotationTest(
        annotation=abi.AssetTransferTransaction,
        expected=abi.AssetTransferTransactionTypeSpec(),
    ),
    TypeAnnotationTest(
        annotation=abi.ApplicationCallTransaction,
        expected=abi.ApplicationCallTransactionTypeSpec(),
    ),
    TypeAnnotationTest(annotation=abi.Account, expected=abi.AccountTypeSpec()),
    TypeAnnotationTest(annotation=abi.Asset, expected=abi.AssetTypeSpec()),
    TypeAnnotationTest(annotation=abi.Application, expected=abi.ApplicationTypeSpec()),
]


@pytest.mark.parametrize("annotation, expected", TYPE_ANNOTATION_TEST_CASES)
def test_type_spec_from_annotation(
    annotation: Any, expected: abi.TypeSpec | type[Exception]
):
    if not isinstance(expected, abi.TypeSpec):
        with pytest.raises(expected):
            abi.type_spec_from_annotation(annotation)
        return

    actual = abi.type_spec_from_annotation(annotation)
    assert actual == expected

    new_instance = actual.new_instance()

    annotation_origin = get_origin(annotation)
    if annotation_origin is None:
        # get_origin will return None for annotations without generic args, e.g. `Byte`
        annotation_origin = annotation

    assert isinstance(
        new_instance, annotation_origin
    ), "TypeSpec.new_instance() returns a value that does not match the annotation type"

    assert (
        actual == new_instance.type_spec()
    ), "TypeSpec.new_instance().type_spec() does not match original TypeSpec"


@pytest.mark.parametrize("subclass", bfs_on_inheritance(abi.BaseType))
def test_type_spec_from_annotation_is_exhaustive(subclass: type):
    # This test is to make sure there are no new subclasses of BaseType that type_spec_from_annotation
    # is not aware of.

    if isabstract(subclass):
        # abstract class type annotations should not be supported
        with pytest.raises(TypeError, match=r"^Unknown annotation origin"):
            abi.type_spec_from_annotation(subclass)
        return

    if subclass is pt.abi.NamedTuple:
        with pytest.raises(pt.TealInputError, match=r"^NamedTuple must be subclassed$"):
            abi.type_spec_from_annotation(subclass)
        return

    try:
        # if subclass is not generic, this will succeed
        abi.type_spec_from_annotation(subclass)
    except TypeError as e:
        # if subclass is generic, we should get an error that is NOT "Unknown annotation origin"
        assert "Unknown annotation origin" not in str(e)

    if issubclass(subclass, pt.abi.NamedTuple):
        # ignore NamedTuple subclasses for the following check
        return

    # make sure there is a testcase for this subclass in test_type_spec_from_annotation
    assert any(
        subclass is testcase.annotation or subclass is get_origin(testcase.annotation)
        for testcase in TYPE_ANNOTATION_TEST_CASES
    ), f"Test case for subclass {subclass} is not present in TYPE_ANNOTATION_TEST_CASES"


def test_make():
    actual = abi.make(abi.Tuple2[abi.Uint64, abi.StaticArray[abi.Bool, Literal[8]]])
    expected_type_spec = abi.TupleTypeSpec(
        abi.Uint64TypeSpec(), abi.StaticArrayTypeSpec(abi.BoolTypeSpec(), 8)
    )

    assert actual.type_spec() == expected_type_spec
    assert type(actual) is abi.Tuple2


def test_size_of():
    values = [
        (abi.Uint8, 1),
        (abi.Address, 32),
        (abi.StaticArray[abi.Uint16, Literal[10]], 2 * 10),
        (abi.StaticBytes[Literal[36]], 36),
    ]

    for (t, s) in values:
        assert abi.size_of(t) == s

    with pytest.raises(pt.TealInputError):
        abi.size_of(abi.String)

    with pytest.raises(pt.TealInputError):
        abi.size_of(abi.DynamicBytes)


ABI_TRANSLATION_TEST_CASES = [
    # Test for byte/bool/address/strings
    (algosdk.abi.ByteType(), "byte", abi.ByteTypeSpec(), abi.Byte),
    (algosdk.abi.BoolType(), "bool", abi.BoolTypeSpec(), abi.Bool),
    (
        algosdk.abi.AddressType(),
        "address",
        abi.AddressTypeSpec(),
        abi.Address,
    ),
    (algosdk.abi.StringType(), "string", abi.StringTypeSpec(), abi.String),
    # Test for dynamic array type
    (
        algosdk.abi.ArrayDynamicType(algosdk.abi.UintType(32)),
        "uint32[]",
        abi.DynamicArrayTypeSpec(abi.Uint32TypeSpec()),
        abi.DynamicArray[abi.Uint32],
    ),
    (
        algosdk.abi.ArrayDynamicType(
            algosdk.abi.ArrayDynamicType(algosdk.abi.ByteType())
        ),
        "byte[][]",
        abi.DynamicArrayTypeSpec(abi.DynamicArrayTypeSpec(abi.ByteTypeSpec())),
        abi.DynamicArray[abi.DynamicArray[abi.Byte]],
    ),
    # TODO: Turn these tests on when PyTeal supports ufixed<N>x<M>
    # cf https://github.com/algorandfoundation/ARCs/blob/main/ARCs/arc-0004.md#types
    # (
    #     algosdk.abi.ArrayDynamicType(algosdk.abi.UfixedType(256, 64)),
    #     "ufixed256x64[]",
    #     abi.DynamicArrayTypeSpec(abi.UfixedTypeSpec(256, 64)),
    # ),
    # # Test for static array type
    # (
    #     algosdk.abi.ArrayStaticType(algosdk.abi.UfixedType(128, 10), 100),
    #     "ufixed128x10[100]",
    #     abi.ArrayStaticTypeSpec(abi.UfixedTypeSpec(128, 10), 100),
    # ),
    (
        algosdk.abi.ArrayStaticType(
            algosdk.abi.ArrayStaticType(algosdk.abi.BoolType(), 256),
            100,
        ),
        "bool[256][100]",
        abi.StaticArrayTypeSpec(
            abi.StaticArrayTypeSpec(abi.BoolTypeSpec(), 256),
            100,
        ),
        abi.StaticArray[abi.StaticArray[abi.Bool, Literal[256]], Literal[100]],
    ),
    # Test for tuple
    (algosdk.abi.TupleType([]), "()", abi.TupleTypeSpec(), abi.Tuple0),
    (
        algosdk.abi.TupleType(
            [
                algosdk.abi.UintType(16),
                algosdk.abi.TupleType(
                    [
                        algosdk.abi.ByteType(),
                        algosdk.abi.ArrayStaticType(algosdk.abi.AddressType(), 10),
                    ]
                ),
            ]
        ),
        "(uint16,(byte,address[10]))",
        abi.TupleTypeSpec(
            abi.Uint16TypeSpec(),
            abi.TupleTypeSpec(
                abi.ByteTypeSpec(),
                abi.StaticArrayTypeSpec(abi.AddressTypeSpec(), 10),
            ),
        ),
        abi.Tuple2[
            abi.Uint16,
            abi.Tuple2[
                abi.Byte,
                abi.StaticArray[abi.Address, Literal[10]],
            ],
        ],
    ),
    (
        algosdk.abi.TupleType(
            [
                algosdk.abi.UintType(64),
                algosdk.abi.TupleType(
                    [
                        algosdk.abi.ByteType(),
                        algosdk.abi.ArrayStaticType(algosdk.abi.AddressType(), 10),
                    ]
                ),
                algosdk.abi.TupleType([]),
                algosdk.abi.BoolType(),
            ]
        ),
        "(uint64,(byte,address[10]),(),bool)",
        abi.TupleTypeSpec(
            abi.Uint64TypeSpec(),
            abi.TupleTypeSpec(
                abi.ByteTypeSpec(),
                abi.StaticArrayTypeSpec(abi.AddressTypeSpec(), 10),
            ),
            abi.TupleTypeSpec(),
            abi.BoolTypeSpec(),
        ),
        abi.Tuple4[
            abi.Uint64,
            abi.Tuple2[
                abi.Byte,
                abi.StaticArray[abi.Address, Literal[10]],
            ],
            abi.Tuple,
            abi.Bool,
        ],
    ),
    # TODO: Turn the following test on when PyTeal supports ufixed<N>x<M>
    # cf https://github.com/algorandfoundation/ARCs/blob/main/ARCs/arc-0004.md#types
    # (
    #     algosdk.abi.TupleType(
    #         [
    #             algosdk.abi.UfixedType(256, 16),
    #             algosdk.abi.TupleType(
    #                 [
    #                     algosdk.abi.TupleType(
    #                         [
    #                             algosdk.abi.StringType(),
    #                         ]
    #                     ),
    #                     algosdk.abi.BoolType(),
    #                     algosdk.abi.TupleType(
    #                         [
    #                             algosdk.abi.AddressType(),
    #                             algosdk.abi.UintType(8),
    #                         ]
    #                     ),
    #                 ]
    #             ),
    #         ]
    #     ),
    #     "(ufixed256x16,((string),bool,(address,uint8)))",
    #     abi.TupleType(
    #         [
    #             abi.UfixedType(256, 16),
    #             abi.TupleType(
    #                 [
    #                     abi.TupleType(
    #                         [
    #                             abi.StringType(),
    #                         ]
    #                     ),
    #                     abi.BoolType(),
    #                     abi.TupleType(
    #                         [
    #                             abi.AddressType(),
    #                             abi.UintType(8),
    #                         ]
    #                     ),
    #                 ]
    #             ),
    #         ]
    #     ),
    # ),
    (
        "cannot map ABI transaction type spec <pyteal.abi.TransactionTypeSpec",
        "txn",
        abi.TransactionTypeSpec(),
        abi.Transaction,
    ),
    (
        "cannot map ABI transaction type spec <pyteal.abi.PaymentTransactionTypeSpec",
        "pay",
        abi.PaymentTransactionTypeSpec(),
        abi.PaymentTransaction,
    ),
    (
        "cannot map ABI transaction type spec <pyteal.abi.KeyRegisterTransactionTypeSpec",
        "keyreg",
        abi.KeyRegisterTransactionTypeSpec(),
        abi.KeyRegisterTransaction,
    ),
    (
        "cannot map ABI transaction type spec <pyteal.abi.AssetConfigTransactionTypeSpec",
        "acfg",
        abi.AssetConfigTransactionTypeSpec(),
        abi.AssetConfigTransaction,
    ),
    (
        "cannot map ABI transaction type spec <pyteal.abi.AssetTransferTransactionTypeSpec",
        "axfer",
        abi.AssetTransferTransactionTypeSpec(),
        abi.AssetTransferTransaction,
    ),
    (
        "cannot map ABI transaction type spec <pyteal.abi.AssetFreezeTransactionTypeSpec",
        "afrz",
        abi.AssetFreezeTransactionTypeSpec(),
        abi.AssetFreezeTransaction,
    ),
    (
        "cannot map ABI transaction type spec <pyteal.abi.ApplicationCallTransactionTypeSpec",
        "appl",
        abi.ApplicationCallTransactionTypeSpec(),
        abi.ApplicationCallTransaction,
    ),
    (
        "cannot map ABI reference type spec <pyteal.abi.AccountTypeSpec",
        "account",
        abi.AccountTypeSpec(),
        abi.Account,
    ),
    (
        "cannot map ABI reference type spec <pyteal.abi.ApplicationTypeSpec",
        "application",
        abi.ApplicationTypeSpec(),
        abi.Application,
    ),
    (
        "cannot map ABI reference type spec <pyteal.abi.AssetTypeSpec",
        "asset",
        abi.AssetTypeSpec(),
        abi.Asset,
    ),
]

ABI_SIGNATURE_TYPESPEC_CASES = [
    (
        "check(uint64,uint64)uint64",
        [abi.Uint64TypeSpec(), abi.Uint64TypeSpec()],
        abi.Uint64TypeSpec(),
    ),
    (
        "check(uint64[],uint64)uint64",
        [abi.DynamicArrayTypeSpec(abi.Uint64TypeSpec()), abi.Uint64TypeSpec()],
        abi.Uint64TypeSpec(),
    ),
    (
        "check(uint64[5],uint64)uint64",
        [abi.StaticArrayTypeSpec(abi.Uint64TypeSpec(), 5), abi.Uint64TypeSpec()],
        abi.Uint64TypeSpec(),
    ),
    (
        "check(uint64,uint64)uint64[]",
        [abi.Uint64TypeSpec(), abi.Uint64TypeSpec()],
        abi.DynamicArrayTypeSpec(abi.Uint64TypeSpec()),
    ),
    (
        "check(uint64,uint64)uint64[5]",
        [abi.Uint64TypeSpec(), abi.Uint64TypeSpec()],
        abi.StaticArrayTypeSpec(abi.Uint64TypeSpec(), 5),
    ),
    (
        "check((uint64,uint64),asset)string",
        [
            abi.TupleTypeSpec(abi.Uint64TypeSpec(), abi.Uint64TypeSpec()),
            abi.AssetTypeSpec(),
        ],
        abi.StringTypeSpec(),
    ),
    (
        "check(string,asset)(uint64,uint64)",
        [abi.StringTypeSpec(), abi.AssetTypeSpec()],
        abi.TupleTypeSpec(abi.Uint64TypeSpec(), abi.Uint64TypeSpec()),
    ),
    (
        "check(account,asset,application)string",
        [abi.AccountTypeSpec(), abi.AssetTypeSpec(), abi.ApplicationTypeSpec()],
        abi.StringTypeSpec(),
    ),
    (
        "check(pay,txn,appl)string",
        [
            abi.PaymentTransactionTypeSpec(),
            abi.TransactionTypeSpec(),
            abi.ApplicationCallTransactionTypeSpec(),
        ],
        abi.StringTypeSpec(),
    ),
    ("check(uint64,uint64)void", [abi.Uint64TypeSpec(), abi.Uint64TypeSpec()], None),
]


@pytest.mark.parametrize(
    "algosdk_abi, abi_string, pyteal_abi_ts, pyteal_abi",
    ABI_TRANSLATION_TEST_CASES,
)
def test_abi_type_translation(algosdk_abi, abi_string, pyteal_abi_ts, pyteal_abi):
    print(f"({algosdk_abi}, {abi_string}, {pyteal_abi_ts}),")

    assert pyteal_abi_ts == abi.type_spec_from_annotation(pyteal_abi)

    assert str(pyteal_abi_ts.new_instance()) == abi_string

    if abi_string in (
        "account",
        "application",
        "asset",
        "txn",
        "pay",
        "keyreg",
        "acfg",
        "axfer",
        "afrz",
        "appl",
    ):
        assert str(pyteal_abi_ts) == abi_string

        with pytest.raises(pt.TealInputError) as tie:
            abi.algosdk_from_type_spec(pyteal_abi_ts)
        assert str(tie.value).startswith(algosdk_abi)

        with pytest.raises(pt.TealInputError) as tie:
            abi.algosdk_from_annotation(pyteal_abi)
        assert str(tie.value).startswith(algosdk_abi)

        return

    assert str(algosdk_abi) == abi_string == str(pyteal_abi_ts)
    assert (
        algosdk_abi
        == algosdk.abi.ABIType.from_string(abi_string)
        == algosdk.abi.ABIType.from_string(str(pyteal_abi_ts))
    )
    assert algosdk_abi == abi.algosdk_from_type_spec(pyteal_abi_ts)
    assert algosdk_abi == abi.algosdk_from_annotation(pyteal_abi)


@pytest.mark.parametrize("case", ABI_TRANSLATION_TEST_CASES)
def test_sdk_abi_translation(case):
    # Errors are strings in the 0th element
    if type(case[0]) is str:
        return
    assert type_spec_from_algosdk(case[0]) == case[2]


@pytest.mark.parametrize("sig_str, sig_args, sig_rets", ABI_SIGNATURE_TYPESPEC_CASES)
def test_sdk_type_specs_from_signature(sig_str, sig_args, sig_rets):
    args, ret = abi.type_specs_from_signature(sig_str)
    assert args == sig_args
    assert ret == sig_rets


class NamedTDecl(abi.NamedTuple):
    a: abi.Field[abi.Uint64]
    b: abi.Field[
        abi.Tuple3[abi.PaymentTransaction, abi.Address, abi.StaticBytes[Literal[16]]]
    ]
    c: abi.Field[abi.Transaction]


class NamedTComp0(abi.NamedTuple):
    a0: abi.Field[abi.String]
    a1: abi.Field[abi.Address]


class NamedTComp1(abi.NamedTuple):
    b0: abi.Field[abi.DynamicBytes]
    b1: abi.Field[abi.StaticBytes[Literal[32]]]


class NamedTComp2(abi.NamedTuple):
    b1: abi.Field[abi.Address]
    b0: abi.Field[abi.DynamicBytes]


class NamedTComp2SameField(abi.NamedTuple):
    b1: abi.Field[abi.Address]
    b0: abi.Field[abi.DynamicBytes]


class SafeBidirectional(NamedTuple):
    xs: list[abi.TypeSpec]


SAFE_BIDIRECTIONAL_TEST_CASES: list[SafeBidirectional] = (
    [
        SafeBidirectional(
            [
                abi.type_spec_from_annotation(abi.StaticArray[abi.Byte, Literal[10]]),
                abi.type_spec_from_annotation(abi.StaticBytes[Literal[10]]),
            ],
        ),
        SafeBidirectional(
            [abi.DynamicArrayTypeSpec(abi.Uint8TypeSpec()), abi.DynamicBytesTypeSpec()]
        ),
        SafeBidirectional([abi.Uint8TypeSpec(), abi.ByteTypeSpec()]),
        SafeBidirectional([abi.BoolTypeSpec()]),
        SafeBidirectional([abi.StringTypeSpec()]),
        SafeBidirectional([abi.AddressTypeSpec()]),
        SafeBidirectional(
            [
                abi.type_spec_from_annotation(abi.DynamicBytes),
                abi.type_spec_from_annotation(abi.DynamicArray[abi.Byte]),
            ]
        ),
        SafeBidirectional(
            [
                abi.type_spec_from_annotation(
                    abi.Tuple3[
                        abi.Uint64,
                        abi.Tuple3[
                            abi.PaymentTransaction,
                            abi.Address,
                            abi.StaticArray[abi.Byte, Literal[16]],
                        ],
                        abi.Transaction,
                    ]
                ),
                abi.type_spec_from_annotation(NamedTDecl),
            ]
        ),
    ]
    + [
        SafeBidirectional([spec])
        for spec in abi.ReferenceTypeSpecs + abi.TransactionTypeSpecs
    ]
    + [
        SafeBidirectional([spec_t()])
        for spec_t in bfs_on_inheritance(abi.UintTypeSpec)
        if not isabstract(spec_t)
    ]
)


@pytest.mark.parametrize("tc", SAFE_BIDIRECTIONAL_TEST_CASES)
def test_type_spec_is_assignable_safe_bidirectional(tc: SafeBidirectional):
    assert len(tc.xs) > 0
    for a in tc.xs:
        for b in tc.xs:
            assert type_spec_is_assignable_to(a, b)


@pytest.mark.parametrize("ts", bfs_on_inheritance(abi.TypeSpec))
def test_type_spec_is_assignable_safe_bidirectional_full_coverage(ts: type):
    def exists_in_safe_bidirectional(_ts: type):
        for safe_bidirectional in SAFE_BIDIRECTIONAL_TEST_CASES:
            for t in safe_bidirectional.xs:
                if type(t) == _ts:
                    return True
        return False

    if isabstract(ts):
        assert not exists_in_safe_bidirectional(ts)
    else:
        assert exists_in_safe_bidirectional(ts)


class SafeAssignment(NamedTuple):
    a: abi.TypeSpec
    bs: list[abi.TypeSpec]


SAFE_ASSIGNMENT_TEST_CASES: list[SafeAssignment] = [
    SafeAssignment(
        abi.StringTypeSpec(),
        [abi.DynamicBytesTypeSpec(), abi.DynamicArrayTypeSpec(abi.ByteTypeSpec())],
    ),
    SafeAssignment(
        abi.type_spec_from_annotation(NamedTDecl),
        [
            abi.type_spec_from_annotation(
                abi.Tuple3[
                    abi.Uint64,
                    abi.Tuple3[
                        abi.Transaction,
                        abi.Address,
                        abi.StaticArray[abi.Byte, Literal[16]],
                    ],
                    abi.Transaction,
                ]
            ),
        ],
    ),
    SafeAssignment(
        abi.type_spec_from_annotation(
            abi.Tuple3[
                abi.Uint64,
                abi.Tuple3[
                    abi.PaymentTransaction,
                    abi.Address,
                    abi.StaticArray[abi.Byte, Literal[16]],
                ],
                abi.PaymentTransaction,
            ]
        ),
        [abi.type_spec_from_annotation(NamedTDecl)],
    ),
    SafeAssignment(
        abi.AddressTypeSpec(),
        [abi.StaticArrayTypeSpec(abi.ByteTypeSpec(), 32), abi.StaticBytesTypeSpec(32)],
    ),
] + [
    SafeAssignment(spec, [abi.TransactionTypeSpec()])
    for spec in abi.TransactionTypeSpecs
    if spec != abi.TransactionTypeSpec()
]


@pytest.mark.parametrize("tc", SAFE_ASSIGNMENT_TEST_CASES)
def test_type_spec_is_assignable_safe_assignment(tc: SafeAssignment):
    assert len(tc.bs) > 0
    for b in tc.bs:
        assert type_spec_is_assignable_to(tc.a, b)
        assert not type_spec_is_assignable_to(b, tc.a)


@pytest.mark.parametrize("ts", bfs_on_inheritance(abi.TypeSpec))
def test_type_spec_is_assignable_safe_assignment_full_coverage(ts: type):
    def exists_in_safe_assignment(_ts: type):
        for safe_assignment in SAFE_ASSIGNMENT_TEST_CASES:
            if type(safe_assignment.a) == _ts:
                return True
            for t in safe_assignment.bs:
                if type(t) == _ts:
                    return True
        return False

    # Abstract types and types without safe assignments should _not_ appear in test cases.
    # These typespecs are only assignable to themselves, and that forms a bidirectional assignment.
    # Otherwise, they are unsafe bidirectional with all the other non-abstract typespecs.
    if isabstract(ts) or ts in {
        abi.AccountTypeSpec,
        abi.ApplicationTypeSpec,
        abi.AssetTypeSpec,
        abi.BoolTypeSpec,
        abi.ByteTypeSpec,
        abi.Uint8TypeSpec,
        abi.Uint16TypeSpec,
        abi.Uint32TypeSpec,
        abi.Uint64TypeSpec,
    }:
        assert not exists_in_safe_assignment(ts)
    else:
        assert exists_in_safe_assignment(ts)


class UnsafeBidirectional(NamedTuple):
    xs: list[abi.TypeSpec]


UNSAFE_BIDIRECTIONAL_TEST_CASES: list[UnsafeBidirectional] = [
    UnsafeBidirectional(
        [
            abi.Uint8TypeSpec(),
            abi.Uint16TypeSpec(),
            abi.Uint32TypeSpec(),
            abi.Uint64TypeSpec(),
            abi.BoolTypeSpec(),
        ]
    ),
    UnsafeBidirectional(
        [
            abi.ByteTypeSpec(),
            abi.Uint16TypeSpec(),
            abi.Uint32TypeSpec(),
            abi.Uint64TypeSpec(),
            abi.BoolTypeSpec(),
        ]
    ),
    UnsafeBidirectional(
        [abi.DynamicBytesTypeSpec(), abi.DynamicArrayTypeSpec(abi.Uint16TypeSpec())]
    ),
    UnsafeBidirectional(
        [abi.StaticBytesTypeSpec(7), abi.StaticArrayTypeSpec(abi.ByteTypeSpec(), 11)]
    ),
    UnsafeBidirectional(
        [
            abi.StringTypeSpec(),
            abi.DynamicArrayTypeSpec(abi.Uint32TypeSpec()),
            abi.DynamicArrayTypeSpec(abi.Uint64TypeSpec()),
            abi.AddressTypeSpec(),
            abi.StaticBytesTypeSpec(33),
        ]
    ),
    UnsafeBidirectional(
        [
            abi.type_spec_from_annotation(NamedTDecl),
            abi.type_spec_from_annotation(NamedTComp0),
            abi.type_spec_from_annotation(NamedTComp1),
            abi.type_spec_from_annotation(NamedTComp2),
            abi.type_spec_from_annotation(NamedTComp2SameField),
        ]
    ),
    UnsafeBidirectional(
        [
            abi.type_spec_from_annotation(NamedTComp1),
            abi.TupleTypeSpec(
                abi.AddressTypeSpec(),
                abi.StaticArrayTypeSpec(abi.Uint16TypeSpec(), 100),
            ),
        ]
    ),
    UnsafeBidirectional(
        [spec for spec in abi.TransactionTypeSpecs if spec != abi.TransactionTypeSpec()]
        + [abi.BoolTypeSpec()]
    ),
    UnsafeBidirectional(abi.ReferenceTypeSpecs + [abi.TransactionTypeSpec()]),
]


@pytest.mark.parametrize("tc", UNSAFE_BIDIRECTIONAL_TEST_CASES)
def test_type_spec_is_assignable_unsafe_bidirectional(tc: UnsafeBidirectional):
    for ia, a in enumerate(tc.xs):
        for ib, b in enumerate(tc.xs):
            if ia == ib:
                assert type_spec_is_assignable_to(a, b)
                assert type_spec_is_assignable_to(b, a)
                continue
            assert not type_spec_is_assignable_to(a, b)


@pytest.mark.parametrize("ts", bfs_on_inheritance(abi.TypeSpec))
def test_type_spec_is_assignable_unsafe_bidirectional_full_coverage(ts: type):
    def exists_in_unsafe_bidirectional(_ts: type):
        for unsafe_bidirectional in UNSAFE_BIDIRECTIONAL_TEST_CASES:
            for t in unsafe_bidirectional.xs:
                if type(t) == _ts:
                    return True
        return False

    if isabstract(ts):
        assert not exists_in_unsafe_bidirectional(ts)
    else:
        assert exists_in_unsafe_bidirectional(ts)



# This example is provided for informational purposes only and has not been audited for security.

from pyteal import *


def approval_program():
    on_creation = Seq(
        [
            Assert(Txn.application_args.length() == Int(1)),
            App.globalPut(Bytes("total supply"), Btoi(Txn.application_args[0])),
            App.globalPut(Bytes("reserve"), Btoi(Txn.application_args[0])),
            App.localPut(Int(0), Bytes("admin"), Int(1)),
            App.localPut(Int(0), Bytes("balance"), Int(0)),
            Return(Int(1)),
        ]
    )

    is_admin = App.localGet(Int(0), Bytes("admin"))

    on_closeout = Seq(
        [
            App.globalPut(
                Bytes("reserve"),
                App.globalGet(Bytes("reserve"))
                + App.localGet(Int(0), Bytes("balance")),
            ),
            Return(Int(1)),
        ]
    )

    register = Seq([App.localPut(Int(0), Bytes("balance"), Int(0)), Return(Int(1))])

    # configure the admin status of the account Txn.accounts[1]
    # sender must be admin
    new_admin_status = Btoi(Txn.application_args[1])
    set_admin = Seq(
        [
            Assert(And(is_admin, Txn.application_args.length() == Int(2))),
            App.localPut(Int(1), Bytes("admin"), new_admin_status),
            Return(Int(1)),
        ]
    )
    # NOTE: The above set_admin code is carefully constructed. If instead we used the following code:
    # Seq([
    #     Assert(Txn.application_args.length() == Int(2)),
    #     App.localPut(Int(1), Bytes("admin"), new_admin_status),
    #     Return(is_admin)
    # ])
    # It would be vulnerable to the following attack: a sender passes in their own address as
    # Txn.accounts[1], so then the line App.localPut(Int(1), Bytes("admin"), new_admin_status)
    # changes the sender's admin status, meaning the final Return(is_admin) can return anything the
    # sender wants. This allows anyone to become an admin!

    # move assets from the reserve to Txn.accounts[1]
    # sender must be admin
    mint_amount = Btoi(Txn.application_args[1])
    mint = Seq(
        [
            Assert(Txn.application_args.length() == Int(2)),
            Assert(mint_amount <= App.globalGet(Bytes("reserve"))),
            App.globalPut(
                Bytes("reserve"), App.globalGet(Bytes("reserve")) - mint_amount
            ),
            App.localPut(
                Int(1),
                Bytes("balance"),
                App.localGet(Int(1), Bytes("balance")) + mint_amount,
            ),
            Return(is_admin),
        ]
    )

    # transfer assets from the sender to Txn.accounts[1]
    transfer_amount = Btoi(Txn.application_args[1])
    transfer = Seq(
        [
            Assert(Txn.application_args.length() == Int(2)),
            Assert(transfer_amount <= App.localGet(Int(0), Bytes("balance"))),
            App.localPut(
                Int(0),
                Bytes("balance"),
                App.localGet(Int(0), Bytes("balance")) - transfer_amount,
            ),
            App.localPut(
                Int(1),
                Bytes("balance"),
                App.localGet(Int(1), Bytes("balance")) + transfer_amount,
            ),
            Return(Int(1)),
        ]
    )

    program = Cond(
        [Txn.application_id() == Int(0), on_creation],
        [Txn.on_completion() == OnComplete.DeleteApplication, Return(is_admin)],
        [Txn.on_completion() == OnComplete.UpdateApplication, Return(is_admin)],
        [Txn.on_completion() == OnComplete.CloseOut, on_closeout],
        [Txn.on_completion() == OnComplete.OptIn, register],
        [Txn.application_args[0] == Bytes("set admin"), set_admin],
        [Txn.application_args[0] == Bytes("mint"), mint],
        [Txn.application_args[0] == Bytes("transfer"), transfer],
    )

    return program


def clear_state_program():
    program = Seq(
        [
            App.globalPut(
                Bytes("reserve"),
                App.globalGet(Bytes("reserve"))
                + App.localGet(Int(0), Bytes("balance")),
            ),
            Return(Int(1)),
        ]
    )

    return program


if __name__ == "__main__":
    with open("asset_approval.teal", "w") as f:
        compiled = compileTeal(approval_program(), mode=Mode.Application, version=2)
        f.write(compiled)

    with open("asset_clear_state.teal", "w") as f:
        compiled = compileTeal(clear_state_program(), mode=Mode.Application, version=2)
        f.write(compiled)



from dataclasses import dataclass, field, fields, astuple
from typing import cast, Optional, Callable
from enum import IntFlag

from algosdk import abi as sdk_abi
from algosdk import encoding

from pyteal.config import METHOD_ARG_NUM_CUTOFF
from pyteal.errors import (
    TealInputError,
    TealInternalError,
)
from pyteal.types import TealType
from pyteal.compiler.compiler import compileTeal, DEFAULT_TEAL_VERSION, OptimizeOptions
from pyteal.ir.ops import Mode

from pyteal.ast import abi
from pyteal.ast.subroutine import (
    OutputKwArgInfo,
    SubroutineFnWrapper,
    ABIReturnSubroutine,
)
from pyteal.ast.assert_ import Assert
from pyteal.ast.cond import Cond
from pyteal.ast.expr import Expr
from pyteal.ast.app import OnComplete
from pyteal.ast.int import Int, EnumInt
from pyteal.ast.seq import Seq
from pyteal.ast.methodsig import MethodSignature
from pyteal.ast.naryexpr import And, Or
from pyteal.ast.txn import Txn
from pyteal.ast.return_ import Approve, Reject


class CallConfig(IntFlag):
    """
    CallConfig: a "bitset"-like class for more fine-grained control over
    `call or create` for a method about an OnComplete case.

    This enumeration class allows for specifying one of the four following cases:
    - CALL
    - CREATE
    - ALL
    - NEVER
    for a method call on one on_complete case.
    """

    NEVER = 0
    CALL = 1
    CREATE = 2
    ALL = 3

    def approval_condition_under_config(self) -> Expr | int:
        match self:
            case CallConfig.NEVER:
                return 0
            case CallConfig.CALL:
                return Txn.application_id() != Int(0)
            case CallConfig.CREATE:
                return Txn.application_id() == Int(0)
            case CallConfig.ALL:
                return 1
            case _:
                raise TealInternalError(f"unexpected CallConfig {self}")

    def clear_state_condition_under_config(self) -> int:
        match self:
            case CallConfig.NEVER:
                return 0
            case CallConfig.CALL:
                return 1
            case CallConfig.CREATE | CallConfig.ALL:
                raise TealInputError(
                    "Only CallConfig.CALL or CallConfig.NEVER are valid for a clear state CallConfig, since clear state can never be invoked during creation"
                )
            case _:
                raise TealInputError(f"unexpected CallConfig {self}")


CallConfig.__module__ = "pyteal"


@dataclass(frozen=True)
class MethodConfig:
    """
    MethodConfig keep track of one method's CallConfigs for all OnComplete cases.

    The `MethodConfig` implementation generalized contract method call such that the registered
    method call is paired with certain OnCompletion conditions and creation conditions.
    """

    no_op: CallConfig = field(kw_only=True, default=CallConfig.NEVER)
    opt_in: CallConfig = field(kw_only=True, default=CallConfig.NEVER)
    close_out: CallConfig = field(kw_only=True, default=CallConfig.NEVER)
    clear_state: CallConfig = field(kw_only=True, default=CallConfig.NEVER)
    update_application: CallConfig = field(kw_only=True, default=CallConfig.NEVER)
    delete_application: CallConfig = field(kw_only=True, default=CallConfig.NEVER)

    def is_never(self) -> bool:
        return all(map(lambda cc: cc == CallConfig.NEVER, astuple(self)))

    def approval_cond(self) -> Expr | int:
        config_oc_pairs: list[tuple[CallConfig, EnumInt]] = [
            (self.no_op, OnComplete.NoOp),
            (self.opt_in, OnComplete.OptIn),
            (self.close_out, OnComplete.CloseOut),
            (self.update_application, OnComplete.UpdateApplication),
            (self.delete_application, OnComplete.DeleteApplication),
        ]
        if all(config == CallConfig.NEVER for config, _ in config_oc_pairs):
            return 0
        elif all(config == CallConfig.ALL for config, _ in config_oc_pairs):
            return 1
        else:
            cond_list = []
            for config, oc in config_oc_pairs:
                config_cond = config.approval_condition_under_config()
                match config_cond:
                    case Expr():
                        cond_list.append(And(Txn.on_completion() == oc, config_cond))
                    case 1:
                        cond_list.append(Txn.on_completion() == oc)
                    case 0:
                        continue
                    case _:
                        raise TealInternalError(
                            f"unexpected condition_under_config: {config_cond}"
                        )
            return Or(*cond_list)

    def clear_state_cond(self) -> Expr | int:
        return self.clear_state.clear_state_condition_under_config()


@dataclass(frozen=True)
class OnCompleteAction:
    """
    OnComplete Action, registers bare calls to one single OnCompletion case.
    """

    action: Optional[Expr | SubroutineFnWrapper | ABIReturnSubroutine] = field(
        kw_only=True, default=None
    )
    call_config: CallConfig = field(kw_only=True, default=CallConfig.NEVER)

    def __post_init__(self):
        if bool(self.call_config) ^ bool(self.action):
            raise TealInputError(
                f"action {self.action} and call_config {self.call_config!r} contradicts"
            )

    @staticmethod
    def never() -> "OnCompleteAction":
        return OnCompleteAction()

    @staticmethod
    def create_only(
        f: Expr | SubroutineFnWrapper | ABIReturnSubroutine,
    ) -> "OnCompleteAction":
        return OnCompleteAction(action=f, call_config=CallConfig.CREATE)

    @staticmethod
    def call_only(
        f: Expr | SubroutineFnWrapper | ABIReturnSubroutine,
    ) -> "OnCompleteAction":
        return OnCompleteAction(action=f, call_config=CallConfig.CALL)

    @staticmethod
    def always(
        f: Expr | SubroutineFnWrapper | ABIReturnSubroutine,
    ) -> "OnCompleteAction":
        return OnCompleteAction(action=f, call_config=CallConfig.ALL)

    def is_empty(self) -> bool:
        return not self.action and self.call_config == CallConfig.NEVER


OnCompleteAction.__module__ = "pyteal"


@dataclass(frozen=True)
class BareCallActions:
    """
    BareCallActions keep track of bare-call registrations to all OnCompletion cases.
    """

    close_out: OnCompleteAction = field(kw_only=True, default=OnCompleteAction.never())
    clear_state: OnCompleteAction = field(
        kw_only=True, default=OnCompleteAction.never()
    )
    delete_application: OnCompleteAction = field(
        kw_only=True, default=OnCompleteAction.never()
    )
    no_op: OnCompleteAction = field(kw_only=True, default=OnCompleteAction.never())
    opt_in: OnCompleteAction = field(kw_only=True, default=OnCompleteAction.never())
    update_application: OnCompleteAction = field(
        kw_only=True, default=OnCompleteAction.never()
    )

    def is_empty(self) -> bool:
        for action_field in fields(self):
            action: OnCompleteAction = getattr(self, action_field.name)
            if not action.is_empty():
                return False
        return True

    def approval_construction(self) -> Optional[Expr]:
        oc_action_pair: list[tuple[EnumInt, OnCompleteAction]] = [
            (OnComplete.NoOp, self.no_op),
            (OnComplete.OptIn, self.opt_in),
            (OnComplete.CloseOut, self.close_out),
            (OnComplete.UpdateApplication, self.update_application),
            (OnComplete.DeleteApplication, self.delete_application),
        ]
        if all(oca.is_empty() for _, oca in oc_action_pair):
            return None
        conditions_n_branches: list[CondNode] = list()
        for oc, oca in oc_action_pair:
            if oca.is_empty():
                continue
            wrapped_handler = ASTBuilder.wrap_handler(
                False,
                cast(Expr | SubroutineFnWrapper | ABIReturnSubroutine, oca.action),
            )
            match oca.call_config:
                case CallConfig.ALL:
                    cond_body = wrapped_handler
                case CallConfig.CALL | CallConfig.CREATE:
                    cond_body = Seq(
                        Assert(
                            cast(
                                Expr, oca.call_config.approval_condition_under_config()
                            )
                        ),
                        wrapped_handler,
                    )
                case _:
                    raise TealInternalError(
                        f"Unexpected CallConfig: {oca.call_config!r}"
                    )
            conditions_n_branches.append(
                CondNode(
                    Txn.on_completion() == oc,
                    cond_body,
                )
            )
        return Cond(*[[n.condition, n.branch] for n in conditions_n_branches])

    def clear_state_construction(self) -> Optional[Expr]:
        if self.clear_state.is_empty():
            return None

        # call this to make sure we error if the CallConfig is CREATE or ALL
        self.clear_state.call_config.clear_state_condition_under_config()

        return ASTBuilder.wrap_handler(
            False,
            cast(
                Expr | SubroutineFnWrapper | ABIReturnSubroutine,
                self.clear_state.action,
            ),
        )


BareCallActions.__module__ = "pyteal"


@dataclass(frozen=True)
class CondNode:
    condition: Expr
    branch: Expr


CondNode.__module__ = "pyteal"


@dataclass
class ASTBuilder:
    conditions_n_branches: list[CondNode] = field(default_factory=list)

    @staticmethod
    def wrap_handler(
        is_method_call: bool, handler: ABIReturnSubroutine | SubroutineFnWrapper | Expr
    ) -> Expr:
        """This is a helper function that handles transaction arguments passing in bare-app-call/abi-method handlers.

        If `is_method_call` is True, then it can only be `ABIReturnSubroutine`,
        otherwise:
            - both `ABIReturnSubroutine` and `Subroutine` takes 0 argument on the stack.
            - all three cases have none (or void) type.

        On ABI method case, if the ABI method has more than 15 args, this function manages to de-tuple
        the last (16-th) Txn app-arg into a list of ABI method arguments, and pass in to the ABI method.

        Args:
            is_method_call: a boolean value that specify if the handler is an ABI method.
            handler: an `ABIReturnSubroutine`, or `SubroutineFnWrapper` (for `Subroutine` case), or an `Expr`.
        Returns:
            Expr:
                - for bare-appcall it returns an expression that the handler takes no txn arg and Approve
                - for abi-method it returns the txn args correctly decomposed into ABI variables,
                  passed in ABIReturnSubroutine and logged, then approve.
        """
        if not is_method_call:
            match handler:
                case Expr():
                    if handler.type_of() != TealType.none:
                        raise TealInputError(
                            f"bare appcall handler should be TealType.none not {handler.type_of()}."
                        )
                    return handler if handler.has_return() else Seq(handler, Approve())
                case SubroutineFnWrapper():
                    if handler.type_of() != TealType.none:
                        raise TealInputError(
                            f"subroutine call should be returning TealType.none not {handler.type_of()}."
                        )
                    if handler.subroutine.argument_count() != 0:
                        raise TealInputError(
                            f"subroutine call should take 0 arg for bare-app call. "
                            f"this subroutine takes {handler.subroutine.argument_count()}."
                        )
                    return Seq(handler(), Approve())
                case ABIReturnSubroutine():
                    if handler.type_of() != "void":
                        raise TealInputError(
                            f"abi-returning subroutine call should be returning void not {handler.type_of()}."
                        )
                    if handler.subroutine.argument_count() != 0:
                        raise TealInputError(
                            f"abi-returning subroutine call should take 0 arg for bare-app call. "
                            f"this abi-returning subroutine takes {handler.subroutine.argument_count()}."
                        )
                    return Seq(cast(Expr, handler()), Approve())
                case _:
                    raise TealInputError(
                        "bare appcall can only accept: none type Expr, or Subroutine/ABIReturnSubroutine with none return and no arg"
                    )
        else:
            if not isinstance(handler, ABIReturnSubroutine):
                raise TealInputError(
                    f"method call should be only registering ABIReturnSubroutine, got {type(handler)}."
                )
            if not handler.is_abi_routable():
                raise TealInputError(
                    f"method call ABIReturnSubroutine is not routable "
                    f"got {handler.subroutine.argument_count()} args with {len(handler.subroutine.abi_args)} ABI args."
                )

            # All subroutine args types
            arg_type_specs = cast(
                list[abi.TypeSpec], handler.subroutine.expected_arg_types
            )

            # All subroutine arg values, initialize here and use below instead of
            # creating new instances on the fly so we dont have to think about splicing
            # back in the transaction types
            arg_vals = [typespec.new_instance() for typespec in arg_type_specs]

            # Only args that appear in app args
            app_arg_vals: list[abi.BaseType] = [
                ats for ats in arg_vals if not isinstance(ats, abi.Transaction)
            ]

            for aav in app_arg_vals:
                # If we're here we know the top level isnt a Transaction but a transaction may
                # be included in some collection type like a Tuple or Array, raise error
                # as these are not supported
                if abi.contains_type_spec(aav.type_spec(), abi.TransactionTypeSpecs):
                    raise TealInputError(
                        "A Transaction type may not be included in Tuples or Arrays"
                    )

            # assign to a var here since we modify app_arg_vals later
            tuplify = len(app_arg_vals) > METHOD_ARG_NUM_CUTOFF

            # only transaction args (these are omitted from app args)
            txn_arg_vals: list[abi.Transaction] = [
                ats for ats in arg_vals if isinstance(ats, abi.Transaction)
            ]

            # Tuple-ify any app args after the limit
            if tuplify:
                tupled_app_args = app_arg_vals[METHOD_ARG_NUM_CUTOFF - 1 :]
                last_arg_specs_grouped: list[abi.TypeSpec] = [
                    t.type_spec() for t in tupled_app_args
                ]
                app_arg_vals = app_arg_vals[: METHOD_ARG_NUM_CUTOFF - 1]
                app_arg_vals.append(
                    abi.TupleTypeSpec(*last_arg_specs_grouped).new_instance()
                )

            # decode app args
            decode_instructions: list[Expr] = [
                app_arg.decode(Txn.application_args[idx + 1])
                for idx, app_arg in enumerate(app_arg_vals)
            ]

            # "decode" transaction types by setting the relative index
            if len(txn_arg_vals) > 0:
                txn_arg_len = len(txn_arg_vals)
                # The transactions should appear in the group in the order they're specified in the method signature
                # and should be relative to the current transaction.

                # ex:
                # doit(axfer,pay,appl)
                # would be 4 transactions
                #      current_idx-3 = axfer
                #      current_idx-2 = pay
                #      current_idx-1 = appl
                #      current_idx-0 = the txn that triggered the current eval (not specified but here for completeness)

                # since we're iterating in order of the txns appearance in the args we
                # subtract the current index from the total length to get the offset.
                # and subtract that from the current index to get the absolute position
                # in the group

                txn_decode_instructions: list[Expr] = []

                for idx, arg_val in enumerate(txn_arg_vals):
                    txn_decode_instructions.append(
                        arg_val._set_index(Txn.group_index() - Int(txn_arg_len - idx))
                    )
                    spec = arg_val.type_spec()
                    if type(spec) is not abi.TransactionTypeSpec:
                        # this is a specific transaction type
                        txn_decode_instructions.append(
                            Assert(arg_val.get().type_enum() == spec.txn_type_enum())
                        )

                decode_instructions += txn_decode_instructions

            # de-tuple into specific values using `store_into` on
            # each element of the tuple'd arguments
            if tuplify:
                tupled_arg: abi.Tuple = cast(abi.Tuple, app_arg_vals[-1])
                de_tuple_instructions: list[Expr] = [
                    tupled_arg[idx].store_into(arg_val)
                    for idx, arg_val in enumerate(tupled_app_args)
                ]
                decode_instructions += de_tuple_instructions

            # NOTE: does not have to have return, can be void method
            if handler.type_of() == "void":
                return Seq(
                    *decode_instructions,
                    cast(Expr, handler(*arg_vals)),
                    Approve(),
                )
            else:
                output_temp: abi.BaseType = cast(
                    OutputKwArgInfo, handler.output_kwarg_info
                ).abi_type.new_instance()
                subroutine_call: abi.ReturnedValue = cast(
                    abi.ReturnedValue, handler(*arg_vals)
                )
                return Seq(
                    *decode_instructions,
                    subroutine_call.store_into(output_temp),
                    abi.MethodReturn(output_temp),
                    Approve(),
                )

    def add_method_to_ast(
        self, method_signature: str, cond: Expr | int, handler: ABIReturnSubroutine
    ) -> None:
        walk_in_cond = Txn.application_args[0] == MethodSignature(method_signature)
        match cond:
            case Expr():
                self.conditions_n_branches.append(
                    CondNode(
                        walk_in_cond,
                        Seq(Assert(cond), self.wrap_handler(True, handler)),
                    )
                )
            case 1:
                self.conditions_n_branches.append(
                    CondNode(walk_in_cond, self.wrap_handler(True, handler))
                )
            case 0:
                return
            case _:
                raise TealInputError("Invalid condition input for add_method_to_ast")

    def program_construction(self) -> Expr:
        if not self.conditions_n_branches:
            return Reject()
        return Cond(*[[n.condition, n.branch] for n in self.conditions_n_branches])


class Router:
    """
    The Router class helps construct the approval and clear state programs for an ARC-4 compliant
    application.

    Additionally, this class can produce an ARC-4 contract description object for the application.

    **WARNING:** The ABI Router is still taking shape and is subject to backwards incompatible changes.

    * Based on feedback, the API and usage patterns are likely to change.
    * Expect migration issues in future PyTeal versions.

    For these reasons, we strongly recommend using :any:`pragma` to pin the version of PyTeal in your
    source code.
    """

    def __init__(
        self,
        name: str,
        bare_calls: BareCallActions = None,
        descr: str = None,
    ) -> None:
        """
        Args:
            name: the name of the smart contract, used in the JSON object.
            bare_calls: the bare app call registered for each on_completion.
            descr: a description of the smart contract, used in the JSON object.
        """

        self.name: str = name
        self.descr = descr

        self.approval_ast = ASTBuilder()
        self.clear_state_ast = ASTBuilder()

        self.methods: list[sdk_abi.Method] = []
        self.method_sig_to_selector: dict[str, bytes] = dict()
        self.method_selector_to_sig: dict[bytes, str] = dict()

        if bare_calls and not bare_calls.is_empty():
            bare_call_approval = bare_calls.approval_construction()
            if bare_call_approval:
                self.approval_ast.conditions_n_branches.append(
                    CondNode(
                        Txn.application_args.length() == Int(0),
                        cast(Expr, bare_call_approval),
                    )
                )
            bare_call_clear = bare_calls.clear_state_construction()
            if bare_call_clear:
                self.clear_state_ast.conditions_n_branches.append(
                    CondNode(
                        Txn.application_args.length() == Int(0),
                        cast(Expr, bare_call_clear),
                    )
                )

    def add_method_handler(
        self,
        method_call: ABIReturnSubroutine,
        overriding_name: str = None,
        method_config: MethodConfig = None,
        description: str = None,
    ) -> ABIReturnSubroutine:
        """Add a method call handler to this Router.

        Args:
            method_call: An ABIReturnSubroutine that implements the method body.
            overriding_name (optional): A name for this method. Defaults to the function name of
                method_call.
            method_config (optional): An object describing the on completion actions and
                creation/non-creation call statuses that are valid for calling this method. All
                invalid configurations will be rejected. Defaults to :code:`MethodConfig(no_op=CallConfig.CALL)`
                (i.e. only the no-op action during a non-creation call is accepted) if none is provided.
            description (optional): A description for this method. Defaults to the docstring of
                method_call, if there is one.
        """
        if not isinstance(method_call, ABIReturnSubroutine):
            raise TealInputError(
                "for adding method handler, must be ABIReturnSubroutine"
            )
        method_signature = method_call.method_signature(overriding_name)
        if method_config is None:
            method_config = MethodConfig(no_op=CallConfig.CALL)
        if method_config.is_never():
            raise TealInputError(
                f"registered method {method_signature} is never executed"
            )
        method_selector = encoding.checksum(bytes(method_signature, "utf-8"))[:4]

        if method_signature in self.method_sig_to_selector:
            raise TealInputError(f"re-registering method {method_signature} detected")
        if method_selector in self.method_selector_to_sig:
            raise TealInputError(
                f"re-registering method {method_signature} has hash collision "
                f"with {self.method_selector_to_sig[method_selector]}"
            )

        meth = method_call.method_spec()
        if description is not None:
            meth.desc = description
        self.methods.append(meth)

        self.method_sig_to_selector[method_signature] = method_selector
        self.method_selector_to_sig[method_selector] = method_signature

        method_approval_cond = method_config.approval_cond()
        method_clear_state_cond = method_config.clear_state_cond()
        self.approval_ast.add_method_to_ast(
            method_signature, method_approval_cond, method_call
        )
        self.clear_state_ast.add_method_to_ast(
            method_signature, method_clear_state_cond, method_call
        )
        return method_call

    def method(
        self,
        func: Callable = None,
        /,
        *,
        name: str = None,
        description: str = None,
        no_op: CallConfig = None,
        opt_in: CallConfig = None,
        close_out: CallConfig = None,
        clear_state: CallConfig = None,
        update_application: CallConfig = None,
        delete_application: CallConfig = None,
    ):
        """This is an alternative way to register a method, as supposed to :code:`add_method_handler`.

        This is a decorator that's meant to be used over a Python function, which is internally
        wrapped with ABIReturnSubroutine. Additional keyword arguments on this decorator can be used
        to specify the OnCompletion statuses that are valid for the registered method.

        NOTE: By default, all OnCompletion actions other than `no_op` are set to `CallConfig.NEVER`,
        while `no_op` field is set to `CallConfig.CALL`. However, if you provide any keywords for
        OnCompletion actions, then the `no_op` field will default to `CallConfig.NEVER`.

        Args:
            func: A function that implements the method body. This should *NOT* be wrapped with the
                :code:`ABIReturnSubroutine` decorator yet.
            name (optional): A name for this method. Defaults to the function name of func.
            description (optional): A description for this method. Defaults to the docstring of
                func, if there is one.
            no_op (optional): The allowed calls during :code:`OnComplete.NoOp`.
            opt_in (optional): The allowed calls during :code:`OnComplete.OptIn`.
            close_out (optional): The allowed calls during :code:`OnComplete.CloseOut`.
            clear_state (optional): The allowed calls during :code:`OnComplete.ClearState`.
            update_application (optional): The allowed calls during :code:`OnComplete.UpdateApplication`.
            delete_application (optional): The allowed calls during :code:`OnComplete.DeleteApplication`.
        """
        # we use `is None` extensively for CallConfig to distinguish 2 following cases
        # - None
        # - CallConfig.Never
        # both cases evaluate to False in if statement.
        def wrap(_func) -> ABIReturnSubroutine:
            wrapped_subroutine = ABIReturnSubroutine(_func, overriding_name=name)
            call_configs: MethodConfig
            if (
                no_op is None
                and opt_in is None
                and close_out is None
                and clear_state is None
                and update_application is None
                and delete_application is None
            ):
                call_configs = MethodConfig(no_op=CallConfig.CALL)
            else:

                def none_to_never(x: None | CallConfig):
                    return CallConfig.NEVER if x is None else x

                _no_op = none_to_never(no_op)
                _opt_in = none_to_never(opt_in)
                _close_out = none_to_never(close_out)
                _clear_state = none_to_never(clear_state)
                _update_app = none_to_never(update_application)
                _delete_app = none_to_never(delete_application)

                call_configs = MethodConfig(
                    no_op=_no_op,
                    opt_in=_opt_in,
                    close_out=_close_out,
                    clear_state=_clear_state,
                    update_application=_update_app,
                    delete_application=_delete_app,
                )
            return self.add_method_handler(
                wrapped_subroutine, name, call_configs, description
            )

        if not func:
            return wrap
        return wrap(func)

    def contract_construct(self) -> sdk_abi.Contract:
        """A helper function in constructing a `Contract` object.

        It takes out the method spec from approval program methods,
        and constructs an `Contract` object.

        Returns:
            A Python SDK `Contract` object constructed from the registered methods on this router.
        """

        return sdk_abi.Contract(self.name, self.methods, self.descr)

    def build_program(self) -> tuple[Expr, Expr, sdk_abi.Contract]:
        """
        Constructs ASTs for approval and clear-state programs from the registered methods and bare
        app calls in the router, and also generates a Contract object to allow client read and call
        the methods easily.

        Note that if no methods or bare app call actions have been registered to either the approval
        or clear state programs, then that program will reject all transactions.

        Returns:
            A tuple of three objects.

            * approval_program: an AST for approval program
            * clear_state_program: an AST for clear-state program
            * contract: a Python SDK Contract object to allow clients to make off-chain calls
        """
        return (
            self.approval_ast.program_construction(),
            self.clear_state_ast.program_construction(),
            self.contract_construct(),
        )

    def compile_program(
        self,
        *,
        version: int = DEFAULT_TEAL_VERSION,
        assemble_constants: bool = False,
        optimize: OptimizeOptions = None,
    ) -> tuple[str, str, sdk_abi.Contract]:
        """
        Constructs and compiles approval and clear-state programs from the registered methods and
        bare app calls in the router, and also generates a Contract object to allow client read and call
        the methods easily.

        This method combines :any:`Router.build_program` and :any:`compileTeal`.

        Note that if no methods or bare app call actions have been registered to either the approval
        or clear state programs, then that program will reject all transactions.

        Returns:
            A tuple of three objects.

            * approval_program: compiled approval program string
            * clear_state_program: compiled clear-state program string
            * contract: a Python SDK Contract object to allow clients to make off-chain calls
        """
        ap, csp, contract = self.build_program()
        ap_compiled = compileTeal(
            ap,
            Mode.Application,
            version=version,
            assembleConstants=assemble_constants,
            optimize=optimize,
        )
        csp_compiled = compileTeal(
            csp,
            Mode.Application,
            version=version,
            assembleConstants=assemble_constants,
            optimize=optimize,
        )
        return ap_compiled, csp_compiled, contract


Router.__module__ = "pyteal"



from typing import cast, overload, TYPE_CHECKING

from pyteal.types import TealType, valid_base16, valid_base32, valid_base64
from pyteal.util import escapeStr
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.errors import TealInputError
from pyteal.ast.leafexpr import LeafExpr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Bytes(LeafExpr):
    """An expression that represents a byte string."""

    @overload
    def __init__(self, arg1: str | bytes | bytearray) -> None:
        pass

    @overload
    def __init__(self, arg1: str, arg2: str) -> None:
        pass

    def __init__(self, arg1: str | bytes | bytearray, arg2: str = None) -> None:
        """
        __init__(arg1: Union[str, bytes, bytearray]) -> None
        __init__(self, arg1: str, arg2: str) -> None

        Create a new byte string.

        Depending on the encoding, there are different arguments to pass:

        For UTF-8 strings:
            Pass the string as the only argument. For example, ``Bytes("content")``.
        For raw bytes or bytearray objects:
            Pass the bytes or bytearray as the only argument. For example, ``Bytes(b"content")``.
        For base16, base32, or base64 strings:
            Pass the base as the first argument and the string as the second argument. For example,
            ``Bytes("base16", "636F6E74656E74")``, ``Bytes("base32", "ORFDPQ6ARJK")``,
            ``Bytes("base64", "Y29udGVudA==")``.
        Special case for base16:
            The prefix "0x" may be present in a base16 byte string. For example,
            ``Bytes("base16", "0x636F6E74656E74")``.
        """
        super().__init__()
        if arg2 is None:
            if type(arg1) is str:
                self.base = "utf8"
                self.byte_str = escapeStr(arg1)
            elif type(arg1) in (bytes, bytearray):
                self.base = "base16"
                self.byte_str = cast(bytes | bytearray, arg1).hex()
            else:
                raise TealInputError(f"Unknown argument type: {type(arg1)}")
        else:
            if type(arg1) is not str:
                raise TealInputError(f"Unknown type for base: {type(arg1)}")

            if type(arg2) is not str:
                raise TealInputError(f"Unknown type for value: {type(arg2)}")

            self.base = arg1

            if self.base == "base32":
                valid_base32(arg2)
                self.byte_str = arg2
            elif self.base == "base64":
                self.byte_str = arg2
                valid_base64(self.byte_str)
            elif self.base == "base16":
                if arg2.startswith("0x"):
                    self.byte_str = arg2[2:]
                else:
                    self.byte_str = arg2
                valid_base16(self.byte_str)
            else:
                raise TealInputError(
                    f"invalid base {self.base}, need to be base32, base64, or base16."
                )

    def __teal__(self, options: "CompileOptions"):
        if self.base == "utf8":
            payload = self.byte_str
        elif self.base == "base16":
            payload = "0x" + self.byte_str
        else:
            payload = f"{self.base}({self.byte_str})"
        op = TealOp(self, Op.byte, payload)
        return TealBlock.FromOp(options, op)

    def __str__(self):
        return f"({self.base} bytes: {self.byte_str})"

    def type_of(self):
        return TealType.bytes


Bytes.__module__ = "pyteal"



# This example is provided for informational purposes only and has not been audited for security.

from pyteal import *

"""Periodic Payment"""

tmpl_fee = Int(1000)
tmpl_period = Int(50)
tmpl_dur = Int(5000)
tmpl_lease = Bytes("base64", "023sdDE2")
tmpl_amt = Int(2000)
tmpl_rcv = Addr("6ZHGHH5Z5CTPCF5WCESXMGRSVK7QJETR63M3NY5FJCUYDHO57VTCMJOBGY")
tmpl_timeout = Int(30000)


def periodic_payment(
    tmpl_fee=tmpl_fee,
    tmpl_period=tmpl_period,
    tmpl_dur=tmpl_dur,
    tmpl_lease=tmpl_lease,
    tmpl_amt=tmpl_amt,
    tmpl_rcv=tmpl_rcv,
    tmpl_timeout=tmpl_timeout,
):

    periodic_pay_core = And(
        Txn.type_enum() == TxnType.Payment,
        Txn.fee() < tmpl_fee,
        Txn.first_valid() % tmpl_period == Int(0),
        Txn.last_valid() == tmpl_dur + Txn.first_valid(),
        Txn.lease() == tmpl_lease,
    )

    periodic_pay_transfer = And(
        Txn.close_remainder_to() == Global.zero_address(),
        Txn.rekey_to() == Global.zero_address(),
        Txn.receiver() == tmpl_rcv,
        Txn.amount() == tmpl_amt,
    )

    periodic_pay_close = And(
        Txn.close_remainder_to() == tmpl_rcv,
        Txn.rekey_to() == Global.zero_address(),
        Txn.receiver() == Global.zero_address(),
        Txn.first_valid() == tmpl_timeout,
        Txn.amount() == Int(0),
    )

    periodic_pay_escrow = periodic_pay_core.And(
        periodic_pay_transfer.Or(periodic_pay_close)
    )

    return periodic_pay_escrow


if __name__ == "__main__":
    print(compileTeal(periodic_payment(), mode=Mode.Signature, version=2))



from abc import ABC, abstractmethod

from typing import Dict, List, Tuple, Set, Iterator, cast, TYPE_CHECKING

from pyteal.ir.tealop import TealOp, Op
from pyteal.errors import TealCompileError

if TYPE_CHECKING:
    from pyteal.ast import Expr, ScratchSlot
    from pyteal.compiler import CompileOptions
    from pyteal.ir.tealsimpleblock import TealSimpleBlock


class TealBlock(ABC):
    """Represents a basic block of TealComponents in a graph."""

    def __init__(self, ops: List[TealOp]) -> None:
        self.ops = ops
        self.incoming: List[TealBlock] = []

    @abstractmethod
    def getOutgoing(self) -> List["TealBlock"]:
        """Get this block's children blocks, if any."""
        pass

    @abstractmethod
    def replaceOutgoing(self, oldBlock: "TealBlock", newBlock: "TealBlock") -> None:
        """Replace one of this block's child blocks."""
        pass

    def isTerminal(self) -> bool:
        """Check if this block ends the program."""
        for op in self.ops:
            if op.getOp() in (Op.return_, Op.retsub, Op.err):
                return True
        return len(self.getOutgoing()) == 0

    def validateTree(
        self, parent: "TealBlock" = None, visited: List["TealBlock"] = None
    ) -> None:
        """Check that this block and its children have valid parent pointers.

        Args:
            parent (optional): The parent block to this one, if it has one. Defaults to None.
            visited (optional): Used internally to remember blocks that have been visited. Set to None.
        """
        if visited is None:
            # using a list instead of a set as TealBlock is not hashable and PyTEAL programs should be short anyway
            visited = []

        if parent is not None:
            count = 0
            for block in self.incoming:
                if parent is block:
                    count += 1
            assert count == 1

        if all(self is not b for b in visited):
            # if the block was not already visited
            visited.append(self)
            for block in self.getOutgoing():
                block.validateTree(self, visited)

    def addIncoming(
        self, parent: "TealBlock" = None, visited: List["TealBlock"] = None
    ) -> None:
        """Calculate the parent blocks for this block and its children.

        Args:
            parent (optional): The parent block to this one, if it has one. Defaults to None.
            visited (optional): Used internally to remember blocks that have been visited. Set to None.
        """
        if visited is None:
            # using a list instead of a set as TealBlock is not hashable and PyTEAL programs should be short anyway
            visited = []

        if parent is not None and all(parent is not b for b in self.incoming):
            self.incoming.append(parent)

        if all(self is not b for b in visited):
            # if the block was not already visited
            visited.append(self)
            for b in self.getOutgoing():
                b.addIncoming(self, visited)

    def validateSlots(
        self,
        slotsInUse: Set["ScratchSlot"] = None,
        visited: Set[Tuple[int, ...]] = None,
    ) -> List[TealCompileError]:
        if visited is None:
            visited = set()

        if slotsInUse is None:
            slotsInUse = set()

        currentSlotsInUse = set(slotsInUse)
        errors = []

        for op in self.ops:
            if op.getOp() == Op.store:
                for slot in op.getSlots():
                    currentSlotsInUse.add(slot)

            if op.getOp() == Op.load:
                for slot in op.getSlots():
                    if slot not in currentSlotsInUse:
                        e = TealCompileError(
                            "Scratch slot load occurs before store", op.expr
                        )
                        errors.append(e)

        if not self.isTerminal():
            sortedSlots = sorted(slot.id for slot in currentSlotsInUse)
            for block in self.getOutgoing():
                visitedKey = (id(block), *sortedSlots)
                if visitedKey in visited:
                    continue
                visited.add(visitedKey)

                for error in block.validateSlots(currentSlotsInUse, visited):
                    if error not in errors:
                        errors.append(error)

        return errors

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        pass

    @classmethod
    def FromOp(
        cls, options: "CompileOptions", op: TealOp, *args: "Expr"
    ) -> Tuple["TealBlock", "TealSimpleBlock"]:
        """Create a path of blocks from a TealOp and its arguments.

        Returns:
            The starting and ending block of the path that encodes the given TealOp and arguments.
        """
        from pyteal.ir.tealsimpleblock import TealSimpleBlock

        opBlock = TealSimpleBlock([op])

        if len(args) == 0:
            return opBlock, opBlock

        start = None
        prevArgEnd = None
        for i, arg in enumerate(args):
            argStart, argEnd = arg.__teal__(options)
            if i == 0:
                start = argStart
            else:
                cast(TealSimpleBlock, prevArgEnd).setNextBlock(argStart)
            prevArgEnd = argEnd

        cast(TealSimpleBlock, prevArgEnd).setNextBlock(opBlock)

        return cast(TealBlock, start), opBlock

    @classmethod
    def Iterate(cls, start: "TealBlock") -> Iterator["TealBlock"]:
        """Perform a breadth-first search of the graph of blocks starting with start."""
        queue = [start]
        visited = list(queue)

        def is_in_visited(block):
            for v in visited:
                if block is v:
                    return True
            return False

        while len(queue) != 0:
            w = queue.pop(0)
            nextBlocks = w.getOutgoing()
            yield w
            for nextBlock in nextBlocks:
                if not is_in_visited(nextBlock):
                    visited.append(nextBlock)
                    queue.append(nextBlock)

    @classmethod
    def NormalizeBlocks(cls, start: "TealBlock") -> "TealBlock":
        """Minimize the number of blocks in the graph of blocks starting with start by combining
        sequential blocks. This operation does not alter the operations of the graph or the
        functionality of its underlying program, however it does mutate the input graph.

        Returns:
            The new starting point of the altered graph. May be the same or differant than start.
        """
        for block in TealBlock.Iterate(start):
            if len(block.incoming) == 1:
                prev = block.incoming[0]
                prevOutgoing = prev.getOutgoing()
                if len(prevOutgoing) == 1 and prevOutgoing[0] is block:
                    # combine blocks
                    block.ops = prev.ops + block.ops
                    block.incoming = prev.incoming
                    for incoming in prev.incoming:
                        incoming.replaceOutgoing(prev, block)
                    if prev is start:
                        start = block

        for block in TealBlock.Iterate(start):
            if len(block.ops) == 0:
                outgoing = block.getOutgoing()
                if len(outgoing) == 1:
                    # if block has 0 ops and 1 outgoing edge, directly connect every incoming block
                    # to the single outgoing block, thereby removing an unnecessary intermediate
                    # jump to this block
                    outgoingBlock = outgoing[0]
                    for i, incomingBlock in enumerate(outgoingBlock.incoming):
                        if block is incomingBlock:
                            # remove block from incoming of outgoing
                            outgoingBlock.incoming.pop(i)
                            break

                    for prev in block.incoming:
                        prev.replaceOutgoing(block, outgoing[0])
                        if id(prev) not in [id(b) for b in outgoingBlock.incoming]:
                            outgoingBlock.incoming.append(prev)

                    if block is start:
                        start = block

        return start

    @classmethod
    def GetReferencedScratchSlots(cls, start: "TealBlock") -> List["ScratchSlot"]:
        """Get all scratch slot references for the graph starting at this TealBlock.

        Returns:
            A list of ScratchSlots where each element represents a reference to that slot by a
            TealOp in the graph. The order of the list is consistent, and there may be duplicate
            ScratchSlots in the list if the same slot is referenced multiple times.
        """
        slots: List[ScratchSlot] = []

        for block in TealBlock.Iterate(start):
            for op in block.ops:
                slots += op.getSlots()

        return slots

    @classmethod
    def MatchScratchSlotReferences(
        cls, actual: List["ScratchSlot"], expected: List["ScratchSlot"]
    ) -> bool:
        """Determine if there is a mapping between the actual and expected lists of ScratchSlots.

        A mapping is defined as follows:
          * The actual and expected lists must have the same length.
          * For every ScratchSlot referenced by either list:

            * If the slot appears in both lists, it must appear the exact same number of times and at
              the exact same indexes in both lists.

            * If the slot appears only in one list, for each of its appearances in that list, there
              must be a ScratchSlot in the other list that appears the exact same number of times
              and at the exact same indexes.

        Returns:
            True if and only if a mapping as described above exists between actual and expected.
        """
        if len(actual) != len(expected):
            return False

        commonSlots = set(actual) & set(expected)
        mapFromActualToExpected: Dict[ScratchSlot, ScratchSlot] = {
            slot: slot for slot in commonSlots
        }

        for actualSlot, expectedSlot in zip(actual, expected):
            if actualSlot not in mapFromActualToExpected:
                if expectedSlot in mapFromActualToExpected.values():
                    # this value was already seen
                    return False
                mapFromActualToExpected[actualSlot] = expectedSlot
                continue

            if mapFromActualToExpected[actualSlot] != expectedSlot:
                return False

        return True


TealBlock.__module__ = "pyteal"



from enum import Enum
from typing import Callable, Optional, Union, cast, TYPE_CHECKING

from pyteal.types import TealType, require_type
from pyteal.errors import (
    TealInputError,
    TealCompileError,
    verifyFieldVersion,
    verifyProgramVersion,
)
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.leafexpr import LeafExpr
from pyteal.ast.expr import Expr
from pyteal.ast.int import EnumInt
from pyteal.ast.array import Array

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class TxnType:
    """Enum of all possible transaction types."""

    Unknown = EnumInt("unknown")
    Payment = EnumInt("pay")
    KeyRegistration = EnumInt("keyreg")
    AssetConfig = EnumInt("acfg")
    AssetTransfer = EnumInt("axfer")
    AssetFreeze = EnumInt("afrz")
    ApplicationCall = EnumInt("appl")


TxnType.__module__ = "pyteal"


class TxnField(Enum):
    sender = (0, "Sender", TealType.bytes, False, 2)
    fee = (1, "Fee", TealType.uint64, False, 2)
    first_valid = (2, "FirstValid", TealType.uint64, False, 2)
    first_valid_time = (3, "FirstValidTime", TealType.uint64, False, 7)
    last_valid = (4, "LastValid", TealType.uint64, False, 2)
    note = (5, "Note", TealType.bytes, False, 2)
    lease = (6, "Lease", TealType.bytes, False, 2)
    receiver = (7, "Receiver", TealType.bytes, False, 2)
    amount = (8, "Amount", TealType.uint64, False, 2)
    close_remainder_to = (9, "CloseRemainderTo", TealType.bytes, False, 2)
    vote_pk = (10, "VotePK", TealType.bytes, False, 2)
    selection_pk = (11, "SelectionPK", TealType.bytes, False, 2)
    vote_first = (12, "VoteFirst", TealType.uint64, False, 2)
    vote_last = (13, "VoteLast", TealType.uint64, False, 2)
    vote_key_dilution = (14, "VoteKeyDilution", TealType.uint64, False, 2)
    type = (15, "Type", TealType.bytes, False, 2)
    type_enum = (16, "TypeEnum", TealType.uint64, False, 2)
    xfer_asset = (17, "XferAsset", TealType.uint64, False, 2)
    asset_amount = (18, "AssetAmount", TealType.uint64, False, 2)
    asset_sender = (19, "AssetSender", TealType.bytes, False, 2)
    asset_receiver = (20, "AssetReceiver", TealType.bytes, False, 2)
    asset_close_to = (21, "AssetCloseTo", TealType.bytes, False, 2)
    group_index = (22, "GroupIndex", TealType.uint64, False, 2)
    tx_id = (23, "TxID", TealType.bytes, False, 2)
    application_id = (24, "ApplicationID", TealType.uint64, False, 2)
    on_completion = (25, "OnCompletion", TealType.uint64, False, 2)
    application_args = (26, "ApplicationArgs", TealType.bytes, True, 2)
    num_app_args = (27, "NumAppArgs", TealType.uint64, False, 2)
    accounts = (28, "Accounts", TealType.bytes, True, 2)
    num_accounts = (2, "NumAccounts", TealType.uint64, False, 2)
    approval_program = (30, "ApprovalProgram", TealType.bytes, False, 2)
    clear_state_program = (31, "ClearStateProgram", TealType.bytes, False, 2)
    rekey_to = (32, "RekeyTo", TealType.bytes, False, 2)
    config_asset = (33, "ConfigAsset", TealType.uint64, False, 2)
    config_asset_total = (34, "ConfigAssetTotal", TealType.uint64, False, 2)
    config_asset_decimals = (35, "ConfigAssetDecimals", TealType.uint64, False, 2)
    config_asset_default_frozen = (
        36,
        "ConfigAssetDefaultFrozen",
        TealType.uint64,
        False,
        2,
    )
    config_asset_unit_name = (37, "ConfigAssetUnitName", TealType.bytes, False, 2)
    config_asset_name = (38, "ConfigAssetName", TealType.bytes, False, 2)
    config_asset_url = (39, "ConfigAssetURL", TealType.bytes, False, 2)
    config_asset_metadata_hash = (
        40,
        "ConfigAssetMetadataHash",
        TealType.bytes,
        False,
        2,
    )
    config_asset_manager = (41, "ConfigAssetManager", TealType.bytes, False, 2)
    config_asset_reserve = (42, "ConfigAssetReserve", TealType.bytes, False, 2)
    config_asset_freeze = (43, "ConfigAssetFreeze", TealType.bytes, False, 2)
    config_asset_clawback = (44, "ConfigAssetClawback", TealType.bytes, False, 2)
    freeze_asset = (45, "FreezeAsset", TealType.uint64, False, 2)
    freeze_asset_account = (46, "FreezeAssetAccount", TealType.bytes, False, 2)
    freeze_asset_frozen = (47, "FreezeAssetFrozen", TealType.uint64, False, 2)
    assets = (48, "Assets", TealType.uint64, True, 3)
    num_assets = (49, "NumAssets", TealType.uint64, False, 3)
    applications = (50, "Applications", TealType.uint64, True, 3)
    num_applications = (51, "NumApplications", TealType.uint64, False, 3)
    global_num_uints = (52, "GlobalNumUint", TealType.uint64, False, 3)
    global_num_byte_slices = (53, "GlobalNumByteSlice", TealType.uint64, False, 3)
    local_num_uints = (54, "LocalNumUint", TealType.uint64, False, 3)
    local_num_byte_slices = (55, "LocalNumByteSlice", TealType.uint64, False, 3)
    extra_program_pages = (56, "ExtraProgramPages", TealType.uint64, False, 4)
    nonparticipation = (57, "Nonparticipation", TealType.uint64, False, 5)
    logs = (58, "Logs", TealType.bytes, True, 5)
    num_logs = (59, "NumLogs", TealType.uint64, False, 5)
    created_asset_id = (60, "CreatedAssetID", TealType.uint64, False, 5)
    created_application_id = (61, "CreatedApplicationID", TealType.uint64, False, 5)
    last_log = (62, "LastLog", TealType.bytes, False, 6)
    state_proof_pk = (63, "StateProofPK", TealType.bytes, False, 6)
    approval_program_pages = (64, "ApprovalProgramPages", TealType.bytes, True, 7)
    num_approval_program_pages = (
        65,
        "NumApprovalProgramPages",
        TealType.uint64,
        False,
        7,
    )
    clear_state_program_pages = (66, "ClearStateProgramPages", TealType.bytes, True, 7)
    num_clear_state_program_pages = (
        67,
        "NumClearStateProgramPages",
        TealType.uint64,
        False,
        7,
    )

    def __init__(
        self, id: int, name: str, type: TealType, is_array: bool, min_version: int
    ) -> None:
        self.id = id
        self.arg_name = name
        self.ret_type = type
        self.is_array = is_array
        self.min_version = min_version

    def type_of(self) -> TealType:
        return self.ret_type


TxnField.__module__ = "pyteal"


class TxnExpr(LeafExpr):
    """An expression that accesses a transaction field from the current transaction."""

    def __init__(self, op: Op, name: str, field: TxnField) -> None:
        super().__init__()
        if field.is_array:
            raise TealInputError("Unexpected array field: {}".format(field))
        self.op = op
        self.name = name
        self.field = field

    def __str__(self):
        return "({} {})".format(self.name, self.field.arg_name)

    def __teal__(self, options: "CompileOptions"):
        verifyFieldVersion(self.field.arg_name, self.field.min_version, options.version)
        verifyProgramVersion(
            self.op.min_version,
            options.version,
            "Program version too low to use op {}".format(self.op),
        )

        op = TealOp(self, self.op, self.field.arg_name)
        return TealBlock.FromOp(options, op)

    def type_of(self):
        return self.field.type_of()


TxnExpr.__module__ = "pyteal"


class TxnaExpr(LeafExpr):
    """An expression that accesses a transaction array field from the current transaction."""

    @staticmethod
    def __validate_index_or_throw(index: Union[int, Expr]):
        if not isinstance(index, (int, Expr)):
            raise TealInputError(
                f"Invalid index type:  Expected int or Expr, but received {index}."
            )
        if isinstance(index, Expr):
            require_type(index, TealType.uint64)

    def __init__(
        self,
        staticOp: Op,
        dynamicOp: Optional[Op],
        name: str,
        field: TxnField,
        index: Union[int, Expr],
    ) -> None:
        super().__init__()
        if not field.is_array:
            raise TealInputError("Unexpected non-array field: {}".format(field))
        self.__validate_index_or_throw(index)

        self.staticOp = staticOp
        self.dynamicOp = dynamicOp
        self.name = name
        self.field = field
        self.index = index

    def __str__(self):
        return "({} {} {})".format(self.name, self.field.arg_name, self.index)

    def __teal__(self, options: "CompileOptions"):
        verifyFieldVersion(self.field.arg_name, self.field.min_version, options.version)

        opToUse = self.staticOp if type(self.index) is int else self.dynamicOp
        if opToUse is None:
            raise TealCompileError("Dynamic array indexing not supported", self)

        verifyProgramVersion(
            opToUse.min_version,
            options.version,
            "Program version too low to use op {}".format(opToUse),
        )

        if type(self.index) is int:
            op = TealOp(self, opToUse, self.field.arg_name, self.index)
            return TealBlock.FromOp(options, op)

        op = TealOp(self, opToUse, self.field.arg_name)
        return TealBlock.FromOp(options, op, cast(Expr, self.index))

    def type_of(self):
        return self.field.type_of()


TxnaExpr.__module__ = "pyteal"


class TxnExprBuilder:
    def __init__(self, op: Op, name: str):
        self.op = op
        self.name = name

    def __call__(self, field: TxnField) -> TxnExpr:
        return TxnExpr(self.op, self.name, field)


TxnExprBuilder.__module__ = "pyteal"


class TxnaExprBuilder:
    def __init__(self, staticOp: Op, dynamicOp: Optional[Op], name: str):
        self.staticOp = staticOp
        self.dynamicOp = dynamicOp
        self.name = name

    def __call__(self, field: TxnField, index: Union[int, Expr]) -> TxnaExpr:
        return TxnaExpr(self.staticOp, self.dynamicOp, self.name, field, index)


TxnaExprBuilder.__module__ = "pyteal"


class TxnArray(Array):
    """Represents a transaction array field."""

    def __init__(
        self, txnObject: "TxnObject", accessField: TxnField, lengthField: TxnField
    ) -> None:
        self.txnObject = txnObject
        self.accessField = accessField
        self.lengthField = lengthField

    def length(self) -> TxnExpr:
        return self.txnObject.makeTxnExpr(self.lengthField)

    def __getitem__(self, index: Union[int, Expr]) -> TxnaExpr:
        if type(index) is int:
            if index < 0:
                raise TealInputError("Invalid array index: {}".format(index))
        else:
            require_type(cast(Expr, index), TealType.uint64)

        return self.txnObject.makeTxnaExpr(self.accessField, index)


TxnArray.__module__ = "pyteal"


class TxnObject:
    """Represents a transaction and its fields."""

    def __init__(
        self,
        makeTxnExpr: Callable[[TxnField], TxnExpr],
        makeTxnaExpr: Callable[[TxnField, Union[int, Expr]], TxnaExpr],
    ) -> None:
        self.makeTxnExpr = makeTxnExpr
        self.makeTxnaExpr = makeTxnaExpr

    def sender(self) -> TxnExpr:
        """Get the 32 byte address of the sender.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#sender
        """
        return self.makeTxnExpr(TxnField.sender)

    def fee(self) -> TxnExpr:
        """Get the transaction fee in micro Algos.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#fee
        """
        return self.makeTxnExpr(TxnField.fee)

    def first_valid(self) -> TxnExpr:
        """Get the first valid round number.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#firstvalid
        """
        return self.makeTxnExpr(TxnField.first_valid)

    def first_valid_time(self) -> TxnExpr:
        """Get the UNIX timestamp of block before txn.FirstValid. Fails if negative.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#firstvalidtime
        """
        return self.makeTxnExpr(TxnField.first_valid_time)

    def last_valid(self) -> TxnExpr:
        """Get the last valid round number.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#lastvalid
        """
        return self.makeTxnExpr(TxnField.last_valid)

    def note(self) -> TxnExpr:
        """Get the transaction note.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#note
        """
        return self.makeTxnExpr(TxnField.note)

    def lease(self) -> TxnExpr:
        """Get the transaction lease.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#lease
        """
        return self.makeTxnExpr(TxnField.lease)

    def receiver(self) -> TxnExpr:
        """Get the 32 byte address of the receiver.

        Only set when :any:`type_enum()` is :any:`TxnType.Payment`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#receiver
        """
        return self.makeTxnExpr(TxnField.receiver)

    def amount(self) -> TxnExpr:
        """Get the amount of the transaction in micro Algos.

        Only set when :any:`type_enum()` is :any:`TxnType.Payment`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#amount
        """
        return self.makeTxnExpr(TxnField.amount)

    def close_remainder_to(self) -> TxnExpr:
        """Get the 32 byte address of the CloseRemainderTo field.

        Only set when :any:`type_enum()` is :any:`TxnType.Payment`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#closeremainderto
        """
        return self.makeTxnExpr(TxnField.close_remainder_to)

    def vote_pk(self) -> TxnExpr:
        """Get the root participation public key.

        Only set when :any:`type_enum()` is :any:`TxnType.KeyRegistration`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#votepk
        """
        return self.makeTxnExpr(TxnField.vote_pk)

    def selection_pk(self) -> TxnExpr:
        """Get the VRF public key.

        Only set when :any:`type_enum()` is :any:`TxnType.KeyRegistration`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#selectionpk
        """
        return self.makeTxnExpr(TxnField.selection_pk)

    def vote_first(self) -> TxnExpr:
        """Get the first round that the participation key is valid.

        Only set when :any:`type_enum()` is :any:`TxnType.KeyRegistration`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#votefirst
        """
        return self.makeTxnExpr(TxnField.vote_first)

    def vote_last(self) -> TxnExpr:
        """Get the last round that the participation key is valid.

        Only set when :any:`type_enum()` is :any:`TxnType.KeyRegistration`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#votelast
        """
        return self.makeTxnExpr(TxnField.vote_last)

    def vote_key_dilution(self) -> TxnExpr:
        """Get the dilution for the 2-level participation key.

        Only set when :any:`type_enum()` is :any:`TxnType.KeyRegistration`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#votekeydilution
        """
        return self.makeTxnExpr(TxnField.vote_key_dilution)

    def nonparticipation(self) -> TxnExpr:
        """Marks an account nonparticipating for rewards.

        Only set when :any:`type_enum()` is :any:`TxnType.KeyRegistration`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#nonparticipation

        Requires program version 5 or higher.
        """
        return self.makeTxnExpr(TxnField.nonparticipation)

    def type(self) -> TxnExpr:
        """Get the type of this transaction as a byte string.

        In most cases it is preferable to use :any:`type_enum()` instead.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#type
        """
        return self.makeTxnExpr(TxnField.type)

    def type_enum(self) -> TxnExpr:
        """Get the type of this transaction.

        See :any:`TxnType` for possible values.
        """
        return self.makeTxnExpr(TxnField.type_enum)

    def xfer_asset(self) -> TxnExpr:
        """Get the ID of the asset being transferred.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetTransfer`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#xferasset
        """
        return self.makeTxnExpr(TxnField.xfer_asset)

    def asset_amount(self) -> TxnExpr:
        """Get the amount of the asset being transferred, measured in the asset's units.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetTransfer`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#assetamount
        """
        return self.makeTxnExpr(TxnField.asset_amount)

    def asset_sender(self) -> TxnExpr:
        """Get the 32 byte address of the subject of clawback.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetTransfer`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#assetsender
        """
        return self.makeTxnExpr(TxnField.asset_sender)

    def asset_receiver(self) -> TxnExpr:
        """Get the recipient of the asset transfer.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetTransfer`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#assetreceiver
        """
        return self.makeTxnExpr(TxnField.asset_receiver)

    def asset_close_to(self) -> TxnExpr:
        """Get the closeout address of the asset transfer.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetTransfer`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#assetcloseto
        """
        return self.makeTxnExpr(TxnField.asset_close_to)

    def group_index(self) -> TxnExpr:
        """Get the position of the transaction within the atomic transaction group.

        A stand-alone transaction is implictly element 0 in a group of 1.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#group
        """
        return self.makeTxnExpr(TxnField.group_index)

    def tx_id(self) -> TxnExpr:
        """Get the 32 byte computed ID for the transaction."""
        return self.makeTxnExpr(TxnField.tx_id)

    def application_id(self) -> TxnExpr:
        """Get the application ID from the ApplicationCall portion of the current transaction.

        Only set when :any:`type_enum()` is :any:`TxnType.ApplicationCall`.
        """
        return self.makeTxnExpr(TxnField.application_id)

    def on_completion(self) -> TxnExpr:
        """Get the on completion action from the ApplicationCall portion of the transaction.

        Only set when :any:`type_enum()` is :any:`TxnType.ApplicationCall`.
        """
        return self.makeTxnExpr(TxnField.on_completion)

    def approval_program(self) -> TxnExpr:
        """Get the approval program.

        Only set when :any:`type_enum()` is :any:`TxnType.ApplicationCall`.
        """
        return self.makeTxnExpr(TxnField.approval_program)

    def clear_state_program(self) -> TxnExpr:
        """Get the clear state program.

        Only set when :any:`type_enum()` is :any:`TxnType.ApplicationCall`.
        """
        return self.makeTxnExpr(TxnField.clear_state_program)

    def rekey_to(self) -> TxnExpr:
        """Get the sender's new 32 byte AuthAddr.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#rekeyto
        """
        return self.makeTxnExpr(TxnField.rekey_to)

    def config_asset(self) -> TxnExpr:
        """Get the asset ID in asset config transaction.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetConfig`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#configasset
        """
        return self.makeTxnExpr(TxnField.config_asset)

    def config_asset_total(self) -> TxnExpr:
        """Get the total number of units of this asset created.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetConfig`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#total"""
        return self.makeTxnExpr(TxnField.config_asset_total)

    def config_asset_decimals(self) -> TxnExpr:
        """Get the number of digits to display after the decimal place when displaying the asset.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetConfig`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#decimals
        """
        return self.makeTxnExpr(TxnField.config_asset_decimals)

    def config_asset_default_frozen(self) -> TxnExpr:
        """Check if the asset's slots are frozen by default or not.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetConfig`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#defaultfrozen"""
        return self.makeTxnExpr(TxnField.config_asset_default_frozen)

    def config_asset_unit_name(self) -> TxnExpr:
        """Get the unit name of the asset.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetConfig`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#unitname"""
        return self.makeTxnExpr(TxnField.config_asset_unit_name)

    def config_asset_name(self) -> TxnExpr:
        """Get the asset name.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetConfig`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#assetname"""
        return self.makeTxnExpr(TxnField.config_asset_name)

    def config_asset_url(self) -> TxnExpr:
        """Get the asset URL.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetConfig`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#url"""
        return self.makeTxnExpr(TxnField.config_asset_url)

    def config_asset_metadata_hash(self) -> TxnExpr:
        """Get the 32 byte commitment to some unspecified asset metdata.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetConfig`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#metadatahash
        """
        return self.makeTxnExpr(TxnField.config_asset_metadata_hash)

    def config_asset_manager(self) -> TxnExpr:
        """Get the 32 byte asset manager address.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetConfig`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#manageraddr"""
        return self.makeTxnExpr(TxnField.config_asset_manager)

    def config_asset_reserve(self) -> TxnExpr:
        """Get the 32 byte asset reserve address.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetConfig`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#reserveaddr"""
        return self.makeTxnExpr(TxnField.config_asset_reserve)

    def config_asset_freeze(self) -> TxnExpr:
        """Get the 32 byte asset freeze address.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetConfig`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#freezeaddr"""
        return self.makeTxnExpr(TxnField.config_asset_freeze)

    def config_asset_clawback(self) -> TxnExpr:
        """Get the 32 byte asset clawback address.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetConfig`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#clawbackaddr"""
        return self.makeTxnExpr(TxnField.config_asset_clawback)

    def created_asset_id(self) -> TxnExpr:
        """Get the asset ID allocated by the creation of an ASA.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetConfig` and this is an asset creation transaction.

        Requires program version 5 or higher.

        * v5 - Only works on inner transactions.
        * >= v6 - Works on top-level and inner transactions.
        """
        return self.makeTxnExpr(TxnField.created_asset_id)

    def freeze_asset(self) -> TxnExpr:
        """Get the asset ID being frozen or un-frozen.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetFreeze`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#freezeasset"""
        return self.makeTxnExpr(TxnField.freeze_asset)

    def freeze_asset_account(self) -> TxnExpr:
        """Get the 32 byte address of the account whose asset slot is being frozen or un-frozen.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetFreeze`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#freezeaccount"""
        return self.makeTxnExpr(TxnField.freeze_asset_account)

    def freeze_asset_frozen(self) -> TxnExpr:
        """Get the new frozen value for the asset.

        Only set when :any:`type_enum()` is :any:`TxnType.AssetFreeze`.

        For more information, see https://developer.algorand.org/docs/reference/transactions/#assetfrozen"""
        return self.makeTxnExpr(TxnField.freeze_asset_frozen)

    def global_num_uints(self) -> TxnExpr:
        """Get the schema count of global state integers in an application creation call.

        Only set when :any:`type_enum()` is :any:`TxnType.ApplicationCall` and this is an app creation call.

        Requires program version 3 or higher.
        """
        return self.makeTxnExpr(TxnField.global_num_uints)

    def global_num_byte_slices(self) -> TxnExpr:
        """Get the schema count of global state byte slices in an application creation call.

        Only set when :any:`type_enum()` is :any:`TxnType.ApplicationCall` and this is an app creation call.

        Requires program version 3 or higher.
        """
        return self.makeTxnExpr(TxnField.global_num_byte_slices)

    def local_num_uints(self) -> TxnExpr:
        """Get the schema count of local state integers in an application creation call.

        Only set when :any:`type_enum()` is :any:`TxnType.ApplicationCall` and this is an app creation call.

        Requires program version 3 or higher.
        """
        return self.makeTxnExpr(TxnField.local_num_uints)

    def local_num_byte_slices(self) -> TxnExpr:
        """Get the schema count of local state byte slices in an application creation call.

        Only set when :any:`type_enum()` is :any:`TxnType.ApplicationCall` and this is an app creation call.

        Requires program version 3 or higher.
        """
        return self.makeTxnExpr(TxnField.local_num_byte_slices)

    def extra_program_pages(self) -> TxnExpr:
        """Get the number of additional pages for each of the application's approval and clear state programs.

        1 additional page means 2048 more total bytes, or 1024 for each program.

        Only set when :any:`type_enum()` is :any:`TxnType.ApplicationCall` and this is an app creation call.

        Requires program version 4 or higher.
        """
        return self.makeTxnExpr(TxnField.extra_program_pages)

    def created_application_id(self) -> TxnExpr:
        """Get the application ID allocated by the creation of an application.

        Only set when :any:`type_enum()` is :any:`TxnType.ApplicationCall` and this is an app creation call.

        Requires program version 5 or higher.

        * v5 - Only works on inner transactions.
        * >= v6 - Works on top-level and inner transactions.
        """
        return self.makeTxnExpr(TxnField.created_application_id)

    def last_log(self) -> TxnExpr:
        """A convenience method for getting the last logged message from a transaction.

        Only application calls may log a message. Returns an empty string if no messages were logged.

        Only set when :any:`type_enum()` is :any:`TxnType.ApplicationCall`.

        Requires program version 6 or higher.
        """
        return self.makeTxnExpr(TxnField.last_log)

    def state_proof_pk(self) -> TxnExpr:
        """Get the state proof public key commitment from a transaction.

        Requires program version 6 or higher.
        """
        return self.makeTxnExpr(TxnField.state_proof_pk)

    @property
    def application_args(self) -> TxnArray:
        """Application call arguments array.

        :type: TxnArray
        """
        return TxnArray(self, TxnField.application_args, TxnField.num_app_args)

    @property
    def accounts(self) -> TxnArray:
        """The accounts array in an ApplicationCall transaction.

        :type: TxnArray
        """
        return TxnArray(self, TxnField.accounts, TxnField.num_accounts)

    @property
    def assets(self) -> TxnArray:
        """The foreign asset array in an ApplicationCall transaction.

        :type: TxnArray

        Requires program version 3 or higher.
        """
        return TxnArray(self, TxnField.assets, TxnField.num_assets)

    @property
    def applications(self) -> TxnArray:
        """The applications array in an ApplicationCall transaction.

        :type: TxnArray

        Requires program version 3 or higher.
        """
        return TxnArray(self, TxnField.applications, TxnField.num_applications)

    @property
    def logs(self) -> TxnArray:
        """The log messages emitted by an application call.

        :type: TxnArray

        Requires program version 5 or higher.

        * v5 - Only works on inner transactions.
        * >= v6 - Works on top-level and inner transactions.
        """
        return TxnArray(self, TxnField.logs, TxnField.num_logs)

    @property
    def approval_program_pages(self) -> TxnArray:
        """The approval program pages.

        :type: TxnArray

        Requires program version 7 or higher.
        """
        return TxnArray(
            self, TxnField.approval_program_pages, TxnField.num_approval_program_pages
        )

    @property
    def clear_state_program_pages(self) -> TxnArray:
        """The clear state program pages.

        :type: TxnArray

        Requires program version 7 or higher.
        """
        return TxnArray(
            self,
            TxnField.clear_state_program_pages,
            TxnField.num_clear_state_program_pages,
        )


TxnObject.__module__ = "pyteal"

Txn: TxnObject = TxnObject(
    TxnExprBuilder(Op.txn, "Txn"), TxnaExprBuilder(Op.txna, Op.txnas, "Txna")
)

Txn.__module__ = "pyteal"
